import { Component, OnInit, ViewChild } from '@angular/core';
import { Subject } from 'rxjs';
import { TranslateService } from '@ngx-translate/core';
import { loadModules } from 'esri-loader';

import {
  faSearch,
  faSearchPlus,
  faSearchMinus,
  faArrowLeft,
  faArrowRight,
  faArrowUp,
  faArrowDown,
  faCheck,
  faQuestion,
  faReply,
  faPen,
  faEraser,
  faCircle,
  faVectorSquare,
  faEllipsisV,
  faBars,
  faMapMarkerAlt,
  faSave,
  faPrint,
  faMapMarkedAlt,
  faDoorOpen,
} from '@fortawesome/free-solid-svg-icons';

import { MapComponent } from '../map/map.component';
import { repeatAction, defaultInterval, defaultDelay } from './utils';
import { MainMenuComponent } from './modals/main-menu/main-menu.component';
import { NavigationHelpComponent } from './modals/navigation-help/navigation-help.component';
import { MapScaleErrorComponent } from './modals/map-scale-error/map-scale-error.component';
import { maxScaleForEdition } from '../map/map-defaults';
import { TopographyLoadingComponent } from './modals/topography-loading/topography-loading.component';
import { DrawingService } from '../map/drawing/drawing.service';
import { DrawingShape, DrawingTexture, DrawingMode } from '../map/drawing/utils';
import { TextureSelectorComponent } from './modals/texture-selector/texture-selector.component';
import { MapsMenuComponent } from './modals/maps-menu/maps-menu.component';
import { EsriMap, ABAToEsriMap } from '../map/maps-database/utils';
import { MapsDatabaseService } from '../map/maps-database/maps-database.service';
import { EditionHelpComponent } from './modals/edition-help/edition-help.component';
import { PrinterService } from '../map/printer/printer.service';
import { Router } from '@angular/router';

@Component({
  selector: 'app-editor',
  templateUrl: './editor.component.html',
  styleUrls: ['./editor.component.css']
})
export class EditorComponent implements OnInit {

  // Svg icons for the UI's navigation buttons.
  faSearch = faSearch;
  faZoomPlus = faSearchPlus;
  faZoomMinus = faSearchMinus;
  faArrowLeft = faArrowLeft;
  faArrowRight = faArrowRight;
  faArrowUp = faArrowUp;
  faArrowDown = faArrowDown;
  faCheck = faCheck;
  faHelp = faQuestion;
  faBack = faReply;

  // Svg icons for the UI's edition buttons.
  faEdit = faPen;
  faDelete = faEraser;
  faCircle = faCircle;
  faPolygon = faVectorSquare;
  faGuideline = faEllipsisV;
  faPedestrian = faBars;
  faMarker = faMapMarkerAlt;
  faSave = faSave;
  faPrint = faPrint;
  faMap = faMapMarkedAlt;
  faLeave = faDoorOpen;

  @ViewChild('map') private map: MapComponent;
  @ViewChild('mainMenu') private mainMenuModal: MainMenuComponent;
  @ViewChild('mapsMenu') private mapsMenuModal: MapsMenuComponent;
  @ViewChild('navigationHelp') private navigationHelpModal: NavigationHelpComponent;
  @ViewChild('mapScaleError') private mapScaleErrorModal: MapScaleErrorComponent;
  @ViewChild('topographyLoading') private topographyLoadingModal: TopographyLoadingComponent;
  @ViewChild('textureSelector') private textureSelectionModal: TextureSelectorComponent;
  @ViewChild('editionHelp') private editionHelpModal: EditionHelpComponent;

  private _state: 'navigation' | 'edition' = 'navigation';

  drawingMode: DrawingMode;
  drawingShape: DrawingShape;

  private _searchText = '';
  private searchWidget: __esri.widgetsSearch;
  private searchSuggestionsResults = new Subject<string[]>();
  searchSuggestionsResults$ = this.searchSuggestionsResults.asObservable();

  private precisionLevel = 3;

  private _basemapVisible = true;
  private _topographyVisible = true;
  private _drawingsVisible = true;

  private _visibleTopographicCategory: 'roads' | 'city' = 'roads';

  private openedMap: EsriMap;
  mapId = '';
  mapTitle = '';

  private stopActionRepetition: () => void | undefined = () => {};

  constructor(
    public translate: TranslateService,
    private mapsDBService: MapsDatabaseService,
    private drawingService: DrawingService,
    private printerService: PrinterService,
    private router: Router
  ) { }

  ngOnInit() {
    this.init();

    // Set the showMapSelectionMenu method of the main menu modal to close the main
    // menu and open the maps selection modal.
    this.mainMenuModal.showMapSelectionMenu = () => {
      this.mainMenuModal.open = false;
      this.mapsMenuModal.open = true;
    };

    // Set the mapSelected method of the map selection modal to open the selected
    // map, switch to edition mode and close the modal.
    this.mapsMenuModal.mapSelected = async (id: number) => {
      this.mapsMenuModal.open = false;
      const abaMap = await this.mapsDBService.getMap(id);
      this.openedMap = await ABAToEsriMap(abaMap);
      this.map.setExtent(this.openedMap.extent);
      this.state = 'edition';
    };

    // Set the backToMainMenu method of the maps menu modal to open the main menu modal.
    this.mapsMenuModal.backToMainMenu = () => {
      this.mapsMenuModal.open = false;
      this.mainMenuModal.open = true;
    };

    // Set the selectTexture method of the texture selection modal to set the texture of
    // the drawing service and start drawing with it.
    this.textureSelectionModal.selectTexture = (texture: DrawingTexture) => {
      this.drawingService.drawingTexture = texture;
      this.textureSelectionModal.open = false;
      this.drawingService.draw(this.map);
    };
  }

  /**
   * Initialise the component with a search widget connected to the MapComponent's view.
   */
  private async init() {
    const [SearchWidget] = await loadModules(['esri/widgets/Search']);

    this.searchWidget = new SearchWidget({
      view: this.map.mapView
    });

    this.map.resizeOnChange = true;
  }

  get state() {
    return this._state;
  }

  set state(value: 'navigation' | 'edition') {
    if (value === 'navigation') {
      this._state = value;
      this.map.clearLayersGraphics();
      this.map.navigationEnabled = true;
      this.basemapVisible = true;
      this.mainMenuModal.open = true;

    } else {
      if (this.map.mapView.scale > maxScaleForEdition) {
        this.mapScaleErrorModal.open = true;
      } else {
        this._state = value;
        this.map.navigationEnabled = false;
        const topographyLoaded = this.loadTopography();

        topographyLoaded.then(
          () => {
            if (this.openedMap) {
              this.map.drawingsLayer.addMany(this.openedMap.graphics);
              this.mapId = '' + this.openedMap.uid;
              this.mapTitle = this.openedMap.title;
              this.visibleTopographicCategory = this.openedMap.city ? 'city' : 'roads';
            }
          }
        );
      }
    }
  }

  get searchText() {
    return this._searchText;
  }

  set searchText(text: string) {
    this._searchText = text;

    if (this._searchText.length > 3) {
      this.searchSuggestions();
    } else {
      this.searchSuggestionsResults.next([]);
    }
  }

  get basemapVisible() {
    return this._basemapVisible;
  }

  set basemapVisible(value: boolean) {
    this._basemapVisible = value;
    this.map.basemapHidden = !value;
  }

  get topographyVisible() {
    return this._topographyVisible;
  }

  set topographyVisible(value: boolean) {
    this._topographyVisible = value;
    this.map.topographyHidden = !value;
  }

  get drawingsVisible() {
    return this._drawingsVisible;
  }

  set drawingsVisible(value: boolean) {
    this._drawingsVisible = value;
    this.map.drawingsHidden = !value;
  }

  get visibleTopographicCategory() {
    return this._visibleTopographicCategory;
  }

  set visibleTopographicCategory(category: 'roads' | 'city') {
    this._visibleTopographicCategory = category;
    this.map.setVisibleTopographicCategory(category);
  }

  /**
   * Load the topography for the current extent of the component's map and
   * make it visible.
   */
  private async loadTopography() {
    this.topographyLoadingModal.open = true;
    this.map.topographyHidden = false;
    this.map.drawingsHidden = false;
    await this.map.loadTopography();
    await this.map.setVisibleTopographicCategory('roads');
    this.topographyLoadingModal.open = false;
  }

  /**
   * Get search suggestions for the text currently set in the component's search input element
   * and show them below it.
   */
  private async searchSuggestions() {
    const suggestions = await this.searchWidget.suggest(this.searchText);

    if (suggestions === null) {
      this.searchSuggestionsResults.next([]);
    } else {
      const results = suggestions.results.reduce<string[]>(
        (total, current) => [...total, ...current.results.map((result) => result.text)],
        []
      );
      this.searchSuggestionsResults.next(results);
    }
  }

  /**
   * Search for an address matching the text in the component's search input and move
   * the map's center to the location of the first result.
   *
   * @param searchText The text for the address search.
   */
  searchAndGoTo(searchText: string) {
    const searchResultPromise = this.searchWidget.search(searchText);
    searchResultPromise.then((searchResults) => {
      const searchResultExtent = searchResults.results[0].results[0].extent;
      this.map.mapView.goTo(searchResultExtent);
    });
  }

  /**
   * Repeatedly change the component's map scale by a given offset until the action
   * is stopped by calling the 'stopActionRepetition' method of the object.
   *
   * @param scaleChange The offset to change the map's scale by.
   */
  private repeatMapScaleChange(scaleChange: number) {
    this.stopActionRepetition = repeatAction(
      () => this.map.modifyMapScale(scaleChange),
      defaultInterval,
      defaultDelay
    );
  }

  /**
   * Action to perform when the zoom plus button is pressed down in the navigation
   * toolbar: zoom in on the component's map scale.
   */
  zoomIn() {
    this.repeatMapScaleChange(-this.precisionLevel);
  }

  /**
   * Action to perform when the zoom minus button is pressed down in the navigation
   * toolbar: zoom out on the component's map scale.
   */
  zoomOut() {
    this.repeatMapScaleChange(this.precisionLevel);
  }

  /**
   * Repeatedly move the component's map center by a given offset until the action
   * is stoped by calling the 'stopActionRepetition' method of the object.
   *
   * @param x The offset to move the map's center by on the x-axis.
   * @param y The offset to move the map's center by on the y-axis.
   */
  private repeatMoveMapCenter(x: number, y: number) {
    this.stopActionRepetition = repeatAction(
      () => this.map.moveMapCenter(x, y),
      defaultInterval,
      defaultDelay
    );
  }

  /**
   * Action to perform when the left arrow button is pressed down in the navigation
   * toolbar: move the component's map center to the left.
   */
  moveLeft() {
    this.repeatMoveMapCenter(-this.precisionLevel, 0);
  }

  /**
   * Action to perform when the right arrow button is pressed down in the navigation
   * toolbar: move the component's map center to the right.
   */
  moveRight() {
    this.repeatMoveMapCenter(this.precisionLevel, 0);
  }

  /**
   * Action to perform when the up arrow button is pressed down in the navigation
   * toolbar: move the component's map center up.
   */
  moveUp() {
    this.repeatMoveMapCenter(0, -this.precisionLevel);
  }

  /**
   * Action to perform when the down arrow button is pressed down in the navigation
   * toolbar: move the component's map center down.
   */
  moveDown() {
    this.repeatMoveMapCenter(0, this.precisionLevel);
  }

  /**
   * Action to perform when the range in the navigation bar is changed:
   * change the precision level of the navigation interface to the new value of the range.
   *
   * @param event The Event triggering the method.
   */
  changePrecision(event: Event) {
    this.precisionLevel = parseInt((event.srcElement as HTMLInputElement).value, 10);
  }

  /**
   * Action to perform when the back button is pressed in the navigation or edition toolbar:
   * open the main menu modal.
   */
  openMainMenu() {
    this.mainMenuModal.open = true;
  }

  /**
   * Action to perform when the help button is clicked in the navigation toolbar:
   * open the navigation help modal.
   */
  showNavigationHelp() {
    this.navigationHelpModal.open = true;
  }

  /**
   * Action to perform when the drawings edition button is pressed:
   * start the edition mode in the drawing service.
   */
  editDrawings() {
    this.drawingMode = DrawingMode.EDITING;
    this.drawingService.editGraphics(this.map);
  }

  /**
   * Action to perform when the drawings deletion button is pressed:
   * start the deletion mode in the drawing service.
   */
  deleteDrawings() {
    this.drawingMode = DrawingMode.ERASING;
    this.drawingService.deleteGraphics(this.map);
  }

  /**
   * Action to perform when the circle button is pressed:
   * open the texture selection modal and start drawing circles
   * on the map with the drawing service.
   */
  drawCircles() {
    this.drawingMode = DrawingMode.DRAWING;
    this.drawingShape = DrawingShape.CIRCLE;
    this.drawingService.drawingShape = DrawingShape.CIRCLE;
    this.textureSelectionModal.open = true;
  }

  /**
   * Action to perform when the polygon button is pressed:
   * open the texture selection modal and start drawing polygons
   * on the map with the drawing service.
   */
  drawPolygons() {
    this.drawingMode = DrawingMode.DRAWING;
    this.drawingShape = DrawingShape.POLYGON;
    this.drawingService.drawingShape = DrawingShape.POLYGON;
    this.textureSelectionModal.open = true;
  }

  /**
   * Action to perform when the guiding line button is pressed:
   * start drawing guiding lines on the map with the drawing service.
   */
  drawLines() {
    this.drawingMode = DrawingMode.DRAWING;
    this.drawingShape = DrawingShape.LINE;
    this.drawingService.drawingShape = DrawingShape.LINE;
    this.drawingService.draw(this.map);
  }

  /**
   * Action to perform when the stripes button is pressed:
   * start drawing stripes on the map with the drawing service.
   */
  drawStripes() {
    this.drawingMode = DrawingMode.DRAWING;
    this.drawingShape = DrawingShape.STRIPES;
    this.drawingService.drawingShape = DrawingShape.STRIPES;
    this.drawingService.draw(this.map);
  }

  /**
   * Action to perform when the marker button is pressed:
   * start drawing markers on the map with the drawing service.
   */
  drawMarker() {
    this.drawingMode = DrawingMode.DRAWING;
    this.drawingShape = DrawingShape.MARKER;
    this.drawingService.drawingShape = DrawingShape.MARKER;
    this.drawingService.draw(this.map);
  }

  /**
   * Action to perform when the save button is pressed in the edition bar:
   * save the current map on the maps database server.
   */
  async saveMap() {
    let message = await this.translate.get('top-bar.map-saving.map-exists-confirm').toPromise();

    // Inform the user that maps cannot be overwritten and that a new map will be saved.
    if (this.openedMap && !confirm(message)) {
      return;
    }

    // Ask the user to input a title for the map.
    message = await this.translate.get('top-bar.map-saving.new-title').toPromise();
    const mapTitle = prompt(message);
    if (!mapTitle) {
      if (mapTitle === '') {
        message = await this.translate.get('top-bar.map-saving.title-error').toPromise();
        alert(message);
      }
      return;
    }

    // Save the map on the server.
    try {
      this.openedMap = {
        title: mapTitle,
        city: this.visibleTopographicCategory === 'city' ? 1 : 0,
        height: this.map.mapView.height,
        width: this.map.mapView.width,
        extent: this.map.mapView.extent,
        graphics: this.map.drawingsLayer.graphics.toArray()
      };
      const newUid = await this.mapsDBService.saveMap(this.openedMap);
      this.openedMap.uid = newUid.id;
      this.mapId = '' + newUid.id;
      this.mapTitle = mapTitle;

      message = await this.translate.get('top-bar.map-saving.success').toPromise();
      alert(message);

    } catch (error) {
      message = await this.translate.get('top-bar.map-saving.error').toPromise();
      alert(message);
      console.log('Error while saving map: ', error.message);
    }
  }

  /**
   * Action to perform when the print button is pressed in the edition bar:
   * print the current map.
   */
  async printMap() {
    // If the map being edited hasn't been saved, show an error message and stop.
    const message = await this.translate.get('editor-modals.print-error.content').toPromise();
    if (!this.openedMap || !this.openedMap.uid) {
      alert(message);
      return;
    }

    // Add a white background to the map.
    const [
      GraphicsLayer,
      Graphic,
      Polygon,
      SimpleFillSymbol
    ] = await loadModules([
      'esri/layers/GraphicsLayer',
      'esri/Graphic',
      'esri/geometry/Polygon',
      'esri/symbols/SimpleFillSymbol'
    ]);

    const extent = this.map.mapView.extent.clone();

    const rectangle = new Polygon({
      spatialReference: this.map.mapView.spatialReference.clone(),
      rings: [
        [extent.xmin, extent.ymin],
        [extent.xmax, extent.ymin],
        [extent.xmax, extent.ymax],
        [extent.xmin, extent.ymax]
      ]
    });
    const fillSymbol = new SimpleFillSymbol({
      color: [255, 255, 255, 1]
    });

    const graphic = new Graphic({
      geometry: rectangle,
      symbol: fillSymbol,
      visible: true
    });

    const background = new GraphicsLayer({
      visible: true
    });
    background.add(graphic);

    this.map.mapView.map.add(background, 0);

    const basemapVisible = this.basemapVisible;
    const topographyVisible = this.topographyVisible;
    const drawingsVisible = this.drawingsVisible;

    this.basemapVisible = false;
    this.topographyVisible = true;
    this.drawingsVisible = true;

    // Wait for the visibility of the layers to have changed before printing the map.
    setTimeout(async () => {
      await this.printerService.printMap(
        this.openedMap.uid ? this.openedMap.uid.toString() : '0',
        this.openedMap.title,
        this.map.mapView
      );

      this.basemapVisible = basemapVisible;
      this.topographyVisible = topographyVisible;
      this.drawingsVisible = drawingsVisible;

      this.map.mapView.map.remove(background);
      this.map.mapView.goTo(extent);
    }, 100);
  }

  /**
   * Action to perform when the open touchpad button is pressed:
   * open the touchpad component with the current map if a saved
   * map is opened in the editor.
   */
  async openTouchpad() {
    if (this.openedMap && this.openedMap.uid) {
      this.router.navigate(['/touchpad', this.openedMap.uid]);
    } else {
      const message = await this.translate.get('editor-modals.touchpad-open-error.content').toPromise();
      alert(message);
    }
  }

  /**
   * Action to perform when the help button is pressed in the edition bar:
   * open the edition help modal.
   */
  showEditionHelp() {
    this.editionHelpModal.open = true;
  }
}
