import { Component, OnInit, Input, OnDestroy } from '@angular/core';
import { BehaviorSubject, Subscription, Observable, combineLatest } from 'rxjs';
import { map, tap } from 'rxjs/operators';

import { MapsDatabaseService } from '../../../map/maps-database/maps-database.service';
import { ABAMap } from '../../../map/maps-database/utils';

const nbMapsPerPage = 10;

import {
  faFilter,
  faChevronLeft,
  faChevronRight,
  faEye
} from '@fortawesome/free-solid-svg-icons';

@Component({
  selector: 'app-maps-menu',
  templateUrl: './maps-menu.component.html',
  styleUrls: ['./maps-menu.component.css']
})
export class MapsMenuComponent implements OnInit, OnDestroy {

  faFilter = faFilter;
  faPrevious = faChevronLeft;
  faNext = faChevronRight;
  faEye = faEye;

  _open = false;

  @Input() mapSelected: (id: number) => void;
  @Input() backToMainMenu: () => void;

  filter$ = new BehaviorSubject<string>('');
  pageNumber$ = new BehaviorSubject<number>(0);
  pages$ = new BehaviorSubject<number[]>([]);
  filteredMaps$ = new BehaviorSubject<ABAMap[]>([]);
  mapsToDisplay$ = new BehaviorSubject<ABAMap[]>([]);

  subscriptions: Subscription[] = [];

  constructor(
    private mapsDBService: MapsDatabaseService
  ) { }

  ngOnInit() {
    this.subscriptions.push(
      this.initFilteredMaps().subscribe(),
      this.initPages().subscribe(),
      this.initPageNumber().subscribe(),
      this.initMapsToDisplay().subscribe()
    );
  }

  ngOnDestroy() {
    this.subscriptions.forEach((subscription) => subscription.unsubscribe());
  }

  get open(): boolean {
    return this._open;
  }

  set open(value: boolean) {
    this._open = value;

    /* Reload the maps from the ABAplans server every time the
       modal is shown. */
    if (value) {
      this.subscriptions[0] = combineLatest(
        this.mapsDBService.getMapsList(),
        this.filter$
      ).pipe(
        map(([maps, filter]) => this.filterMaps(maps, filter)),
        tap((maps) => this.filteredMaps$.next(maps))
      ).subscribe();
    }
  }

  int(value: string): number {
    return parseInt(value, 10);
  }

  private filterMaps(maps: ABAMap[], filter: string): ABAMap[] {
    return maps.filter(
      (abaMap) => [
        abaMap.uid.toString(),
        abaMap.title.toLowerCase(),
        abaMap.creationDate.toLowerCase()
      ].some(
        (candidate) => candidate.includes(filter)
      )
    );
  }

  private initFilteredMaps(): Observable<ABAMap[]> {
    return combineLatest(
      this.mapsDBService.getMapsList(),
      this.filter$
    ).pipe(
      map(([maps, filter]) => this.filterMaps(maps, filter)),
      tap((maps) => this.filteredMaps$.next(maps))
    );
  }

  private initPages(): Observable<number[]> {
    return this.filteredMaps$.pipe(
      map((maps) => {
        const numberOfPages = Math.ceil(maps.length / nbMapsPerPage);
        return Array(numberOfPages).fill(null).map(({}, i) => i);
      }),
      tap((pages) => this.pages$.next(pages))
    );
  }

  private initPageNumber(): Observable<any> {
    return this.filteredMaps$.pipe(
      tap(() => this.pageNumber$.next(0)),
    );
  }

  private initMapsToDisplay(): Observable<any> {
    return combineLatest(
      this.filteredMaps$,
      this.pageNumber$,
    ).pipe(
      map(([maps, pageNumber]) => {
        const startOffset = nbMapsPerPage * pageNumber;

        return maps.slice(
          startOffset,
          startOffset + nbMapsPerPage,
        );
      }),
      tap((maps) => this.mapsToDisplay$.next(maps)),
    );
  }

}
