/**
 * Repeat an action at given intervals and return a function to be called to stop the repetition.
 *
 * @param action The function called repeatedly at the given intervals.
 * @param repetitionInterval The interval for the repetitions in ms.
 * @param delay The delay before starting to repeat the action in ms.
 *
 * @returns A function to stop the repetition of the action.
 */
export function repeatAction(action: () => void, repetitionInterval: number, delay: number): () => void {
  let stopRepetition: () => void;

  action();

  const timeoutId = setTimeout(() => {
    const intervalId = setInterval(() => {
      action();
    }, repetitionInterval);

    stopRepetition = () => clearInterval(intervalId);
  }, delay);

  stopRepetition = () => clearTimeout(timeoutId);

  return () => stopRepetition();
}

/**
 * Default interval between repetitions of an action in the interface
 * (such as movement on the map) in ms.
 */
export const defaultInterval = 100;

/**
 * Default delay before an action is triggered in the interface in ms.
 */
export const defaultDelay = 500;
