/**
 * 2D vector.
 */
export interface Vector2d {
  x: number;
  y: number;
}

/**
 * 2D plane.
 */
export interface Plane2d {
  A: Vector2d;
  B: Vector2d;
  C: Vector2d;
  D: Vector2d;
}

/**
 * Origin vector.
 */
export const originVector: Vector2d = {
  x: 0,
  y: 0
};

/**
 * Clone a vector.
 *
 * @param vector The vector to clone.
 *
 * @returns A new instance of Vector2d with the same coordinates.
 */
export function clone(vector: Vector2d): Vector2d {
  return {
    x: vector.x,
    y: vector.y
  };
}

/**
 * Compute the euclidean norm of a 2D vector.
 *
 * @param vector The vector for which the euclidean norm must be computed.
 *
 * @returns The norm of the input vector.
 */
export function norm(vector: Vector2d): number {
  return Math.sqrt(vector.x * vector.x + vector.y * vector.y);
}

/**
 * Get the normal vector of a 2D line (another 2D vector).
 *
 * @param vector The vector for which the normal must be computed.
 *
 * @returns A new 2D vector normal to the input.
 */
export function normal(vector: Vector2d): Vector2d {
  const vectorNorm = norm(vector);
  return {
    x: -vector.y / vectorNorm,
    y: vector.x / vectorNorm
  };
}

/**
 * Add two vectors.
 *
 * @param vector1 A Vector2d.
 * @param vector2 Another Vector2d to add to the first.
 *
 * @returns A new Vector2d corresponding to the addition of the inputs.
 */
export function add(vector1: Vector2d, vector2: Vector2d): Vector2d {
  return {
    x: vector1.x + vector2.x,
    y: vector1.y + vector2.y
  };
}

/**
 * Subtract two vectors.
 *
 * @param vector1 A Vector2d to subtract from.
 * @param vector2 The Vector2d to subtract from the first.
 *
 * @returns A new Vector2d corresponding to the subtraction of the two inputs.
 */
export function subtract(vector1: Vector2d, vector2: Vector2d): Vector2d {
  return {
    x: vector1.x - vector2.x,
    y: vector1.y - vector2.y
  };
}

/**
 * 2D cross product between vectors.
 *
 * @param vector1 The first vector in the product.
 * @param vector2 The second vector in the product.
 *
 * @returns The result of the 2D cross product between the vectors.
 */
export function crossProduct(vector1: Vector2d, vector2: Vector2d): number {
  return vector1.x * vector2.y - vector1.y * vector2.x;
}

/**
 * Multiply a vector by a scalar.
 *
 * @param vector The Vector2d which must be multiplied by a scalar.
 * @param scalar The scalar to multiply the vector by.
 *
 * @returns A new Vector2d corresponding to the result of the multiplication.
 */
export function scalarMultiplication(
  vector: Vector2d,
  scalar: number
): Vector2d {
  return {
    x: vector.x * scalar,
    y: vector.y * scalar
  };
}

/**
 * Divide a vector by a scalar.
 *
 * @param vector The Vector2d which must be divided by a scalar.
 * @param scalar The scalar to divide the vector by.
 *
 * @returns A new Vector2d corresponding to the result of the division.
 */
export function scalarDivision(
  vector: Vector2d,
  scalar: number
): Vector2d {
  return {
    x: vector.x / scalar,
    y: vector.y / scalar
  };
}

/**
 * Convert a Vector2d to an array of numbers.
 *
 * @param vector The Vector2d to convert to an array.
 *
 * @returns An array containing the x and y coordinates of the input vector.
 */
export function vectorToArray(vector: Vector2d): Array<number> {
  return [vector.x, vector.y];
}

/**
 * Convert a an array of 2 numbers to a Vector2d.
 *
 * @param vector The Vector2d to convert to an array.
 *
 * @returns An array containing the x and y coordinates of the input vector.
 */
export function arrayToVector(array: number[]): Vector2d {
  return {
    x: array[0],
    y: array[1]
  };
}

/**
 * Transform coordinates from a device plane to a map plane.
 *
 * We use the barycentric coordinate system to apply linear transformation helping
 * to find the final coordinates in an well scaled plan from a deformed plan
 *
 * Algo:
 * - Divide the plan between two triangles
 * - Find if a point is on the left or on the right of BC
 * - Apply barycentric coordinate system to find alpha, beta and gamma
 *   parameters
 * - Thanks to these parameters, transform the coordinates according
 *   to the well scaled vertices.
 *
 *
 *    A ------ C        A --------- C
 *   /     __/ |        |      ___/ |
 *   |  __/    \    =>  |  ___/     |
 *   / /     x  \       | /      x  |
 *  B -----_____ D      B --------- D
 *
 * @param pressedPoint The point pressed on the device plane.
 * @param devicePlane The device's plane.
 * @param mapPlane The map plane to transform the coordinates to.
 */
export function transformCoordinates(
  pressedPoint: Vector2d,
  devicePlane: Plane2d,
  mapPlane: Plane2d
): Vector2d {
  if (devicePlane.A === undefined || devicePlane.B === undefined ||
      devicePlane.C === undefined || devicePlane.D === undefined) {
    return {x: 0, y: 0} as Vector2d;
  }

  const PA = subtract(devicePlane.A, pressedPoint);
  const PB = subtract(devicePlane.B, pressedPoint);
  const PC = subtract(devicePlane.C, pressedPoint);
  const PD = subtract(devicePlane.D, pressedPoint);

  const BC = subtract(devicePlane.C, devicePlane.B);
  const BP = subtract(pressedPoint, devicePlane.B);

  let PX;
  let mapX;

  if (crossProduct(BP, BC) >= 0) {
    /*
     *  A --------- C
     *  |  P   ___/ |
     *  |  ___/     |
     *  | /         |
     *  B --------- D
     */
    PX = PA;
    mapX = mapPlane.A;
  } else {
    /*
     *  A --------- C
     *  |      ___/ |
     *  |  ___/     |
     *  | /     P   |
     *  B --------- D
     */
    PX = PD;
    mapX = mapPlane.D;
  }

  // Apply barycentric coordinates system to find the main parameters
  const alpha = crossProduct(PB, PC);
  const beta = crossProduct(PC, PX);
  const gamma = crossProduct(PX, PB);

  // transformation. P_ is P'
  const total = alpha + beta + gamma;
  const P_ = add(
      scalarMultiplication(mapX, alpha),
      add(
          scalarMultiplication(mapPlane.B, beta),
          scalarMultiplication(mapPlane.C, gamma)
      )
  );
  P_.x /= total;
  P_.y /= total;

  return P_;
}
