import { Injectable } from '@angular/core';
import { TranslateService } from '@ngx-translate/core';
import { loadModules } from 'esri-loader';

import { MapComponent } from '../map.component';
import { CircleDrawer } from './circle-drawer';
import { PolygonDrawer } from './polygon-drawer';
import { LineDrawer } from './line-drawer';
import { StripesDrawer } from './stripes-drawer';
import { MarkerDrawer } from './marker-drawer';
import { DrawingShape, DrawingTexture, stripesWidth } from './utils';
import { subtract, arrayToVector, scalarMultiplication, add, normal } from '../../geometry/geometry2d';
import { BsModalService } from 'ngx-bootstrap';

@Injectable({
  providedIn: 'root'
})
export class DrawingService {

  drawingShape: DrawingShape;
  drawingTexture: DrawingTexture;

  private editor: __esri.SketchViewModel;

  private actionListeners: IHandle[] = [];

  private circleDrawer = new CircleDrawer();
  private polygonDrawer = new PolygonDrawer();
  private lineDrawer = new LineDrawer();
  private stripesDrawer = new StripesDrawer();
  private markerDrawer: MarkerDrawer;

  constructor(
    private translate: TranslateService,
    private modalService: BsModalService
  ) {
    this.markerDrawer = new MarkerDrawer(this.translate, modalService);
  }

  /**
   * Clear all the service's listeners for DrawAction events.
   */
  clearListeners() {
    if (this.editor) {
      this.editor.complete();
    }

    this.actionListeners.forEach((listener) => listener.remove());
    this.actionListeners = [];

    this.circleDrawer.clearListeners();
    this.polygonDrawer.clearListeners();
    this.lineDrawer.clearListeners();
    this.stripesDrawer.clearListeners();
    this.markerDrawer.clearListeners();
  }

  /**
   * Start drawing the current shape with the current texture on the MapComponent passed
   * as input to the method.
   *
   * @param mapView The MapComponent on which the shape must be drawn.
   */
  async draw(
    map: MapComponent
  ) {
    this.clearListeners();

    if (this.drawingShape === DrawingShape.POLYGON) {
      this.polygonDrawer.draw(map, this.drawingTexture);
    } else if (this.drawingShape === DrawingShape.CIRCLE) {
      this.circleDrawer.draw(map, this.drawingTexture);
    } else if (this.drawingShape === DrawingShape.LINE) {
      this.lineDrawer.draw(map);
    } else if (this.drawingShape === DrawingShape.STRIPES) {
      this.stripesDrawer.draw(map);
    } else if (this.drawingShape === DrawingShape.MARKER) {
      this.markerDrawer.draw(map);
    }
  }

  /**
   * Start a handler to edit graphics in the drawingsLayer of a MapComponent.
   *
   * @param map The MapComponent whose graphics must be edited.
   */
  async editGraphics(
    map: MapComponent
  ) {
    const [
      SketchViewModel
    ] = await loadModules([
      'esri/widgets/Sketch/SketchViewModel'
    ]);

    this.clearListeners();

    this.editor = new SketchViewModel({
      view: map.mapView,
      layer: map.drawingsLayer,
      updateOnGraphicClick: false,
      defaultUpdateOperations: {
        toggleToolOnClick: false
      }
    });

    this.actionListeners.push(
      map.mapView.on(
        'click',
        (event) => {
          this.handleEditionClick(event, map);
        }
      )
    );

    // A listener is added on the 'update' event of the editor to handle special cases
    // such as lines and stripes.
    this.actionListeners.push(

      this.editor.on(
        'update',
        async (event) => {
          if (event.state === 'complete') {
            if (event.graphics[0].geometry.type === 'polyline') {

              // When a line (polyline) is edited, a white background
              // must be added to it again once modifications are over.
              this.editor.layer.remove(event.graphics[0]);
              await this.lineDrawer.setBackground(map, event.graphics[0].attributes.id, event.graphics[0].geometry);
              this.editor.layer.add(event.graphics[0]);

            } else if (event.graphics[0].attributes &&
                       event.graphics[0].attributes.temporaryStripes) {

              // When stripes are edited, the graphics (black and white stripes) that compose it
              // are removed from the maps drawingsLayer and it is replaced by a simple temporary rectangle for edition.
              // Once edition on is complete, the stripes must be added back to the graphics layer,
              // and the temporary replacement rectangle is deleted.
              const line = subtract(
                arrayToVector(event.graphics[0].geometry.rings[0][0]),
                arrayToVector(event.graphics[0].geometry.rings[0][1])
              );
              const lineNormal = scalarMultiplication(normal(line), stripesWidth);

              const start = add(
                arrayToVector(event.graphics[0].geometry.rings[0][0]),
                lineNormal
              );
              const end = add(
                arrayToVector(event.graphics[0].geometry.rings[0][1]),
                lineNormal
              );

              this.editor.layer.remove(event.graphics[0]);
              this.stripesDrawer.createFinalStripes(
                map,
                start,
                end
              );
            }
          }
        }
      )
    );
  }

  /**
   * React to a click on a map to edit graphics.
   *
   * @param event The MapViewClickEvent to react to.
   * @param map The MapComponent on which the graphics to edit are.
   */
  private async handleEditionClick(
    event: __esri.MapViewClickEvent,
    map: MapComponent
  ) {
    if (this.editor.state === 'active') {
      return;
    }

    map.mapView.hitTest(event).then(
      (response) => {
        const result = response.results[0];

        if (result.graphic.layer === this.editor.layer) {
          // Ignore guiding lines backgrounds.
          if (result.graphic.attributes && result.graphic.attributes.background) {
            return;
          }

          /* Guiding lines and pedestrian passages need to be handled specifically, as they
             are composed of multiple graphics instead of one. */
          if (result.graphic.geometry.type === 'polyline') {
            const background = this.lineDrawer.getBackground(map, result.graphic);
            this.editor.layer.graphics.remove(background);
            this.editor.update([result.graphic], {tool: 'transform'});

          // Markers are edited by setting the text associated to them.
          } else if (result.graphic.geometry.type === 'point') {
            this.markerDrawer.setTextAttribute(result.graphic);

          } else if (result.graphic.attributes && result.graphic.attributes.stripes) {
            this.handleStripesEdition(map, result.graphic);

          } else {
            this.editor.update([result.graphic], {tool: 'transform'});
          }
        }
      });
  }

  /**
   * Handle a click on stripes to edit them.
   *
   * @param map The MapComponent on which the stripes that must be edited are displayed.
   * @param graphic The graphic belonging to the pedestrian passage
   * that must be edited and which was clicked.
   */
  private async handleStripesEdition(
    map: MapComponent,
    graphic: __esri.Graphic
  ) {
    /* When a passage is edited, its graphics are retrieved and removed from the graphics layer,
       and it is temporarily replaced by a simple rectangle. */
    const id = graphic.attributes.id;
    const startEnd = this.stripesDrawer.deleteStripes(map, id);
    const temporaryStripes = await this.stripesDrawer.createTemporaryStripes(map.mapView, startEnd[0], startEnd[1]);
    this.editor.layer.add(temporaryStripes);
    this.editor.update([temporaryStripes], {tool: 'transform'});
  }

  /**
   * Start a handler to react to clicks on the input MapComponent to delete graphics.
   *
   * @param map The MapComponent on which the user clicks to delete graphics.
   * @param graphicsLayer The GraphicsLayer in which the graphics to delete are.
   */
  async deleteGraphics(
    map: MapComponent,
  ) {
    this.clearListeners();

    this.actionListeners.push(
      map.mapView.on(
        'click',
        (event: __esri.MapViewClickEvent) => this.handleDeletionClick(event, map)
      )
    );
  }

  /**
   * React to a click on a MapComponent to delete some graphics.
   *
   * @param event The MapViewClickEvent to react to.
   * @param map The MapComponent on which the graphics to delete are.
   */
  private async handleDeletionClick(
    event: __esri.MapViewClickEvent,
    map: MapComponent
  ) {
    const [
      geometryEngine,
      Circle
    ] = await loadModules([
      'esri/geometry/geometryEngine',
      'esri/geometry/Circle'
    ]);

    const matchingGraphics = map.drawingsLayer.graphics.filter(
      // First check if the graphic has a geometry, and then if it intersects with the click.
      (graphic) => graphic.geometry && geometryEngine.intersects(
        graphic.geometry,
        // Create a circle geometry around the position of the click to handle clicks on a line.
        new Circle({center: event.mapPoint, radius: 3}))
    );

    const messagesPromises: Promise<string>[] = [];
    messagesPromises.push(this.translate.get('top-bar.confirmMultipleGraphicsDeletion').toPromise());
    messagesPromises.push(this.translate.get('top-bar.confirmSingleGraphicDeletion').toPromise());

    Promise.all(messagesPromises).then(
      (messages) => {
        if (
          matchingGraphics.length > 1 && confirm(messages[0]) ||
          matchingGraphics.length === 1 && confirm(messages[1])
        ) {
          matchingGraphics.forEach(
            (graphic) => {
              // Stripes are represented by a set of graphics, they hence need to be handled
              // differently (each graphic which is part of it must be retrieved and deleted separately).
              if (graphic.attributes && graphic.attributes.stripes) {
                this.stripesDrawer.deleteStripes(map, graphic.attributes.id);
              } else {
                map.drawingsLayer.remove(graphic);
              }
            }
          );
        }
      }
    );
  }
}
