import { loadModules } from 'esri-loader';

import { Drawer } from './drawer';
import { MapComponent } from '../map.component';
import { DrawActionEvent, getDefaultOutline } from './utils';

export class LineDrawer extends Drawer {

  private lastId = 0;

  /**
   * Draw a line on a map.
   *
   * @param map The MapComponent on which the line must be drawn.
   */
  async draw(
    map: MapComponent
  ) {
    const [Draw] = await loadModules(['esri/views/2d/draw/Draw']);

    this.clearListeners();

    const draw: __esri.Draw = new Draw({
      view: map.mapView
    });
    const drawAction = draw.create('polyline', {});

    this.drawActionListeners.push(
      drawAction.on(
        'vertex-add',
        async (event) => await this.createLine(event, map)
      ),

      drawAction.on(
        'vertex-remove',
        async (event) => await this.createLine(event, map)
      ),

      drawAction.on(
        'cursor-update',
        async (event) => await this.createLine(event, map)
      ),

      drawAction.on(
        'draw-complete',
        async (event) => {
          await this.createLine(event, map);
          this.currentGraphic = null;
          this.draw(map);
        }
      ),
    );
  }

  /**
   * Generate a polyline for the vertices captured by the DrawActionEvent which
   * triggered the line's drawing and add it to the drawingsLayer of a MapComponent.
   * 
   * @param event The DrawActionEvent which triggered the line's drawing.
   * @param map The MapComponent to which the line must be added.
   */
  private async createLine(
    event: DrawActionEvent,
    map: MapComponent
  ) {
    const [
      Graphic,
      Polyline,
      SimpleLineSymbol
    ] = await loadModules([
      'esri/Graphic',
      'esri/geometry/Polyline',
      'esri/symbols/SimpleLineSymbol'
    ]);

    if (this.currentGraphic != null) {
      map.drawingsLayer.remove(this.currentGraphic);
    }

    const polyline: __esri.Polyline = new Polyline({
      spatialReference: map.mapView.spatialReference.clone(),
      paths: event.vertices
    });

    let drawingSymbol: __esri.SimpleLineSymbol;

    if (event.type === 'draw-complete') {
      drawingSymbol = new SimpleLineSymbol({
        color: '#000000',
        width: 2,
        style: 'dash',
      });

      this.lastId++;

      /* Add a white background behind the line to make it visible
         both on black and white surfaces. */
      await this.setBackground(map, this.lastId, polyline);

    } else {
      drawingSymbol = await getDefaultOutline();
    }

    this.currentGraphic = new Graphic({
      geometry: polyline,
      symbol: drawingSymbol,
      attributes: {
        id: this.lastId,
        background: false
      }
    });

    if (this.currentGraphic != null) {
      map.drawingsLayer.add(this.currentGraphic);
    }
  }

  /**
   * Add a white background behind a line.
   *
   * @param map The map on which the line and background are being drawn.
   * @param lineId The id of the line behind which a ackground must be added.
   * @param polyline The geometry of the line behind which a background must be added.
   */
  async setBackground(
    map: MapComponent,
    lineId: number,
    polyline: __esri.Geometry
  ) {
    const [
      SimpleLineSymbol,
      Graphic
    ] = await loadModules([
      'esri/symbols/SimpleLineSymbol',
      'esri/Graphic',
    ]);

    const backgroundSymbol = new SimpleLineSymbol({
      color: '#ffffff',
      width: 8,
      style: 'solid'
    });
    const background = new Graphic({
      geometry: polyline,
      symbol: backgroundSymbol,
      attributes: {
        id: lineId,
        background: true
      }
    });
    map.drawingsLayer.add(background);
  }

  /**
   * Retrieve the background associated to a line.
   *
   * @param map The MapComponent on which the line and background are displayed.
   * @param polyline The line for which the background must be retrieved.
   *
   * @returns The Esri Graphic containing the input line's background.
   */
  getBackground(
    map: MapComponent,
    polyline: __esri.Graphic,
  ): __esri.Graphic {
    let background: __esri.Graphic = {} as __esri.Graphic;
    map.drawingsLayer.graphics.forEach(
      (graphic) => {
        if (graphic.attributes &&
            graphic.attributes.id === polyline.attributes.id &&
            graphic.attributes.background) {
          background = graphic;
        }
      }
    );
    return background;
  }
}
