import { loadModules } from 'esri-loader';
import { TranslateService } from '@ngx-translate/core';
import { BsModalService, BsModalRef } from 'ngx-bootstrap';

import { Drawer } from './drawer';
import { MapComponent } from '../map.component';
import { MarkerDrawActionEvent } from './utils';
import { MarkerPromptComponent } from './marker-prompt/marker-prompt.component';

export class MarkerDrawer extends Drawer {

  bsModalRef: BsModalRef;

  constructor(
    private translate: TranslateService,
    private modalService: BsModalService
  ) {
    super();
  }

  /**
   * Draw a marker on a map.
   *
   * @param map The MapComponent on which the marker must be drawn.
   */
  async draw(map: MapComponent) {
    const [Draw] = await loadModules(['esri/views/2d/draw/Draw']);

    this.clearListeners();

    const draw: __esri.Draw = new Draw({
      view: map.mapView
    });
    const drawAction = draw.create('point', {});

    this.drawActionListeners.push(
      drawAction.on(
        'draw-complete',
        async (event) => {
          await this.createMarker(event, map);
          this.draw(map);
        }
      )
    );
  }

  /**
   * Create a marker (Esri Point) and add it to a map.
   *
   * @param event The MarkerDrawActionEvent which triggered the drawing of the marker.
   * @param map The MapComponent to which the marker must be added.
   */
  private async createMarker(
    event: MarkerDrawActionEvent,
    map: MapComponent
  ) {
    const [
      Graphic,
      Point,
      SimpleMarkerSymbol
    ] = await loadModules([
      'esri/Graphic',
      'esri/geometry/Point',
      'esri/symbols/SimpleMarkerSymbol'
    ]);

    const markerBackgroundSymbol = new SimpleMarkerSymbol({
      color: [0, 0, 0, 1.0],
      size: '28px',
      outline: {
        color: [255, 255, 255, 1.0],
        width: '3px'
      }
    });

    const drawingSymbol = new SimpleMarkerSymbol({
      color: [0, 0, 0, 1.0],
      size: '14px',
      outline: {
        color: [255, 255, 255, 1.0],
        width: '3px'
      }
    });

    const point = new Point({
      spatialReference: map.mapView.spatialReference.clone(),
      x: event.coordinates[0],
      y: event.coordinates[1]
    });

    const markerBackground = new Graphic({
      geometry: point,
      symbol: markerBackgroundSymbol,
      visible: true
    });

    const marker = new Graphic({
      geometry: point,
      symbol: drawingSymbol,
      visible: true,
      attributes: {
        text: ''
      }
    });

    const success = await this.setTextAttribute(marker);

    if (success) {
      map.drawingsLayer.addMany([markerBackground, marker]);
    } else {
      const messagePromise = this.translate.get('top-bar.marker-edition.text-error').toPromise();
      const message = await messagePromise;
      alert(message);
    }
  }

  /**
   * Set the text associated to a marker.
   *
   * @param marker An Esri Graphic representing a marker to which text must be associated.
   *
   * @returns A boolean indicating if the text could be set for the marker or not.
   */
  async setTextAttribute(
    marker: __esri.Graphic
  ): Promise<boolean> {
    const initialText = marker.attributes.text;
    const initialState = {
      confirmed: false,
      text: initialText
    };
    this.bsModalRef = this.modalService.show(MarkerPromptComponent, {initialState});

    return new Promise((resolve, reject) => {
      const waitForConfirmation = () => {
        if (this.bsModalRef.content.confirmed) {
          if (this.bsModalRef.content.text) {
            marker.attributes.text = this.bsModalRef.content.text;
            return resolve(true);
          }
          return resolve(false);
        }
        setTimeout(waitForConfirmation, 100);
      };
      waitForConfirmation();
    });
  }
}
