import { loadModules } from 'esri-loader';

import { Drawer } from './drawer';
import { MapComponent } from '../map.component';
import { DrawingTexture, DrawActionEvent, getDefaultOutline, getDrawingSymbol } from './utils';

export class PolygonDrawer extends Drawer {

  /**
   * Draw a polygon on a map.
   *
   * @param map The MapComponent on which the polygon must be drawn.
   * @param texture The texture which must be used to draw the polygon.
   */
  async draw(
    map: MapComponent,
    texture: DrawingTexture
  ) {
    const [Draw] = await loadModules(['esri/views/2d/draw/Draw']);

    this.clearListeners();

    const draw: __esri.Draw = new Draw({
      view: map.mapView
    });
    const drawAction = draw.create('polygon', {});

    this.drawActionListeners.push(
      drawAction.on(
        'vertex-add',
        async (event) => await this.createPolygon(event, map, texture)
      ),

      drawAction.on(
        'vertex-remove',
        async (event) => await this.createPolygon(event, map, texture)
      ),

      drawAction.on(
        'cursor-update',
        async (event) => await this.createPolygon(event, map, texture)
      ),

      drawAction.on(
        'draw-complete',
        async (event) => {
          await this.createPolygon(event, map, texture);
          this.currentGraphic = null;
          this.draw(map, texture);
        }
      ),
    );
  }

  /**
   * Generate a poylgon for the vertices captured by a DrawActionEvent and
   * add it to the drawingsLayer of a MapComponent.
   *
   * @param event The DrawActionEvent which triggered the polygon's drawing.
   * @param map The MapComponent on which the polygon must be drawn.
   * @param texture The texture to use to draw the polygon.
   */
  private async createPolygon(
    event: DrawActionEvent,
    map: MapComponent,
    texture: DrawingTexture
  ) {
    const [
      Graphic,
      Polygon,
      Polyline,
      SimpleFillSymbol
    ] = await loadModules([
      'esri/Graphic',
      'esri/geometry/Polygon',
      'esri/geometry/Polyline',
      'esri/symbols/SimpleFillSymbol'
    ]);

    const drawingSymbol: __esri.SimpleFillSymbol = new SimpleFillSymbol({
      color: [143, 143, 143, 0.75]
    });

    if (this.currentGraphic != null) {
      map.drawingsLayer.remove(this.currentGraphic);
    }

    if (Array(event.vertices).length === 2) {
      const polyline: __esri.Polyline = new Polyline({
        spatialReference: map.mapView.spatialReference.clone(),
        paths: [Array(event.vertices)]
      });

      const lineSymbol = await getDefaultOutline();

      this.currentGraphic = new Graphic({
        geometry: polyline,
        symbol: lineSymbol
      });

    } else {
      const polygon: __esri.Polygon = new Polygon({
        spatialReference: map.mapView.spatialReference.clone(),
        rings: event.vertices
      });

      this.currentGraphic = new Graphic({
        geometry: polygon,
        symbol: drawingSymbol,
      });

      if (this.currentGraphic != null) {
        if (event.type === 'cursor-update') {
          (this.currentGraphic.symbol as __esri.FillSymbol).outline = await getDefaultOutline();

        } else if (event.type === 'draw-complete') {
          this.currentGraphic.symbol = await getDrawingSymbol(texture);
          // (this.currentGraphic.symbol as __esri.FillSymbol).outline = {} as __esri.SimpleLineSymbol;
        }

        map.drawingsLayer.add(this.currentGraphic);
      }
    }
  }
}
