import { loadModules } from 'esri-loader';

import { Drawer } from './drawer';
import { MapComponent } from '../map.component';
import { DrawActionEvent, getDefaultOutline, stripesWidth } from './utils';
import { Vector2d, clone, add, subtract, norm, normal, scalarMultiplication, vectorToArray } from '../../geometry/geometry2d';

export class StripesDrawer extends Drawer {

  private lastId = 0;

  /**
   * Draw stripes on a map.
   *
   * @param map The MapComponent on which the stripes must be drawn.
   */
  async draw(
    map: MapComponent
  ) {
    const [Draw] = await loadModules(['esri/views/2d/draw/Draw']);

    this.clearListeners();

    const draw: __esri.Draw = new Draw({
      view: map.mapView
    });
    const drawAction = draw.create('rectangle', {});

    this.drawActionListeners.push(
      drawAction.on(
        'vertex-add',
        async (event) => await this.createStripes(event, map)
      ),

      drawAction.on(
        'cursor-update',
        async (event) => await this.createStripes(event, map)
      ),

      drawAction.on(
        'draw-complete',
        async (event) => {
          await this.createStripes(event, map);
          this.currentGraphic = null;
          this.draw(map);
        }
      )
    );
  }

  /**
   * Generate stripes for given vertices captured by a DrawActionEvent and add them
   * to a MapComponent.
   *
   * @param event The DrawActionEvent which captured the vertices for which stripes must be generated.
   * @param map The MapComponent on which the stripes must be added.
   */
  private async createStripes(
    event: DrawActionEvent,
    map: MapComponent
  ) {
    if (this.currentGraphic != null) {
      map.drawingsLayer.remove(this.currentGraphic);
    }

    /* Only start drawing the passage when the user has clicked and is
       dragging the mouse. */
    if (event.vertices.length === 2) {
      const start: Vector2d = {x: event.vertices[0][0], y: event.vertices[0][1]};
      const end: Vector2d = {x: event.vertices[1][0], y: event.vertices[1][1]};

      if (event.type === 'draw-complete') {
        this.createFinalStripes(map, start, end);

      } else {
        this.currentGraphic = await this.createTemporaryStripes(map.mapView, start, end);
        map.drawingsLayer.add(this.currentGraphic);
      }
    }
  }

  /**
   * Create a temporary stripes rectanle to use while drawing or editing stripes on a map.
   *
   * @param mapView The Esri MapView on which the temporary stripes must be drawn.
   * @param start The coordinates of the starting point of the stripes rectangle.
   * @param end The coordinates of the ending point of the stripes rectangle.
   *
   * @returns A Graphic containing the temporary stripes rectangle to display.
   */
  async createTemporaryStripes(
    mapView: __esri.MapView,
    start: Vector2d,
    end: Vector2d
  ): Promise<__esri.Graphic> {
    const [
      Graphic,
      SimpleFillSymbol
    ] = await loadModules([
      'esri/Graphic',
      'esri/symbols/SimpleFillSymbol'
    ]);

    const rectangle = await this.createRectangle(mapView, start, end);

    const drawingSymbol: __esri.Symbol = new SimpleFillSymbol({
      color: [143, 143, 143, 0.75]
    });
    const outline: __esri.SimpleLineSymbol = await getDefaultOutline();

    const temporaryStripes = new Graphic({
      geometry: rectangle,
      symbol: drawingSymbol,
      visible: true,
      attributes: {
        temporaryStripes: true
      }
    });
    (temporaryStripes.symbol as __esri.FillSymbol).outline = outline;

    return temporaryStripes;
  }

  /**
   * Create the Graphics for a stripes drawing and add them to a map.
   *
   * @param map The MapComponent on which the stripes must be drawn.
   * @param start The coordinates of the starting point of the stripes on the map.
   * @param end The coordinates of the ending point of the stripes.
   */
  async createFinalStripes(
    map: MapComponent,
    start: Vector2d,
    end: Vector2d
  ) {
    const [
      Graphic,
      SimpleFillSymbol
    ] = await loadModules([
      'esri/Graphic',
      'esri/symbols/SimpleFillSymbol',
    ]);

    const line = subtract(start, end);
    const lineNorm = norm(line);

    // Set the number of stripes composing the drawing.
    let nbIters = Math.max(3, Math.floor(lineNorm / 2.5));
    nbIters = nbIters % 2 ? nbIters + 1 : nbIters;
    const jump = scalarMultiplication(line, 1 / nbIters);

    /* Create a graphic for each stripe and add it to the graphics layer.
       Stripes belonging to the same drawing will have the same id so they can
       be processed together when editing or deleting stripes. */
    let currentStart = clone(start);
    for (let i = 0; i < nbIters; i++) {
      const currentEnd = subtract(currentStart, jump);

      const rectangle = await this.createRectangle(map.mapView, currentStart, currentEnd);

      const currentSymbol = (i + 1) % 2 ?
        new SimpleFillSymbol({color: '#000000'}) :
        new SimpleFillSymbol({color: '#ffffff'});

      const startVector = i === 0 ? clone(currentStart) : null;
      const endVector = i === nbIters - 1 ? clone(currentEnd) : null;

      const stripe = new Graphic({
        geometry: rectangle,
        symbol: currentSymbol,
        visible: true,
        attributes: {
          id: this.lastId,
          stripes: true,
          start: startVector,
          end: endVector
        }
      });

      map.drawingsLayer.add(stripe);
      currentStart = clone(currentEnd);
    }
    this.lastId++;
  }

  /**
   * Create a rectangle geometry between a start and end point on an Esri MapView.
   *
   * @param mapView The MapView holding the spatial reference to which the
   * rectangle must be associated.
   * @param start The starting point of the rectangle.
   * @param end The ending point of the rectangle.
   *
   * @returns A new poylgon geometry with the shape of a rectangle between
   * the input start and end points.
   */
  private async createRectangle(
    mapView: __esri.MapView,
    start: Vector2d,
    end: Vector2d,
  ): Promise<__esri.Geometry> {
    const [Polygon] = await loadModules(['esri/geometry/Polygon']);

    // Line connecting the point clicked and the current position of the mouse.
    const line = subtract(start, end);

    // Normal vector to the line connecting the points.
    const lineNormal = scalarMultiplication(normal(line), stripesWidth);

    // Vertices of a rectangle between the two points.
    const vertices = [
      subtract(start, lineNormal),
      subtract(end, lineNormal),
      add(end, lineNormal),
      add(start, lineNormal)
    ];

    // Convert the vertices to arrays to pass them to the Polygon constructor.
    const verticesArray = vertices.map((vertex) => vectorToArray(vertex));

    const rectangle: __esri.Geometry = new Polygon({
      spatialReference: mapView.spatialReference.clone(),
      rings: verticesArray
    });

    return rectangle;
  }

  /**
   * Remove stripes from a map and return the coordinates of
   * its starting and ending points.
   *
   * @param map The MapComponent from which the stripes must be removed.
   * @param id The id of the stripes group which must be removed from the map
   *
   * @returns An array of 2 Vector2d instances containing the coordinates of the
   * starting and ending points of the deleted stripes.
   */
  deleteStripes(
    map: MapComponent,
    id: number
  ): Vector2d[] {
    let start: Vector2d = {x: 0, y: 0};
    let end: Vector2d = {x: 0, y: 0};
    const toRemove: __esri.Graphic[] = [];

    map.drawingsLayer.graphics.forEach(
      (graphic) => {
        if (graphic.attributes &&
            graphic.attributes.stripes &&
            graphic.attributes.id === id) {
          if (graphic.attributes.start) {
            start = graphic.attributes.start;
          } else if (graphic.attributes.end) {
            end = graphic.attributes.end;
          }
          toRemove.push(graphic);
        }
      }
    );

    // Only remove graphics after iterating over all of them.
    map.drawingsLayer.removeMany(toRemove);
    return [start, end];
  }
}
