import { loadModules } from 'esri-loader';

import { linesImgUrl, dotImgUrl } from '../../../config/urls';


/**
 * The available modes for drawing in the application.
 */
export enum DrawingMode {
  DRAWING = 'drawing',
  EDITING = 'editing',
  ERASING = 'erasing'
}

/**
 * The shapes available for drawing in the application.
 */
export enum DrawingShape {
  CIRCLE = 'circle',
  POLYGON = 'polygon',
  LINE = 'line',
  STRIPES = 'stripes',
  MARKER = 'marker'
}

/**
 * The textures available for drawing in the application.
 */
export enum DrawingTexture {
  DOTS = 'dots',
  LINES = 'lines',
  BLACK = 'black',
  WHITE = 'white'
}

/**
 * Get the Esri Symbol for a given texture.
 *
 * @param texture The DrawingTexture for which the symbol must be retrieved.
 *
 * @returns The Esri Symbol for the input texture.
 */
export async function getDrawingSymbol(texture: DrawingTexture): Promise<__esri.Symbol> {
  const [
    SimpleFillSymbol,
    PictureFillSymbol
  ] = await loadModules([
    'esri/symbols/SimpleFillSymbol',
    'esri/symbols/PictureFillSymbol'
  ]);

  if (texture === DrawingTexture.BLACK) {
    return new SimpleFillSymbol({
      color: '#000000'
    });
  } else if (texture === DrawingTexture.WHITE) {
    return new SimpleFillSymbol({
      color: '#ffffff'
    });
  } else if (texture === DrawingTexture.DOTS) {
    return new PictureFillSymbol({
      url: dotImgUrl,
      height: '16px',
      width: '16px'
    });
  } else {
    return new PictureFillSymbol({
      url: linesImgUrl,
      height: '16px',
      width: '16px'
    });
  }
}

/**
 * Get a default outline for drawings.
 *
 * @returns A SimpleLineSymbol with a default style for the outline of shapes
 * being drawn on a map.
 */
export async function getDefaultOutline(): Promise<__esri.SimpleLineSymbol> {
  const [SimpleLineSymbol] = await loadModules(['esri/symbols/SimpleLineSymbol']);

  return new SimpleLineSymbol({
    color: '#ff0000',
    width: 2,
    style: 'solid'
  });
}

/**
 * The default width for stripes (pedestrian passages) drawings on maps.
 */
export const stripesWidth = 7;

/**
 * The interface for events triggered when a shape is being drawn in the application.
 */
export interface DrawActionEvent {
  type: string;
  vertices: number[][];
}

/**
 * The interface for events triggered when a marker is being drawn in the application.
 */
export interface MarkerDrawActionEvent {
  type: string;
  coordinates: number[];
}
