import { async, ComponentFixture, TestBed } from '@angular/core/testing';

import { MapComponent } from './map.component';
import { loadModules } from 'esri-loader';
import { getAllTopographyTypes, TopographyTypeName } from './topography/utils';
import { TopographyService } from './topography/topography.service';

class TopographyServiceStub {
  async loadAllTopographyTypesGraphics(extent: __esri.Extent, scale: number): Promise<__esri.GraphicsLayer[]> {
    const [GraphicsLayer] = await loadModules(['esri/layers/GraphicsLayer']);

    const topographyTypes = await getAllTopographyTypes();
    const topographyGraphics: __esri.GraphicsLayer[] = [];
    topographyTypes.forEach(
      (type) => {
        const graphicsLayer: __esri.GraphicsLayer = new GraphicsLayer({
          id: type.name,
          visible: false
        });
        topographyGraphics.push(graphicsLayer);
      }
    );
    return topographyGraphics;
  }
}

describe('MapComponent', () => {
  let component: MapComponent;
  let fixture: ComponentFixture<MapComponent>;

  beforeEach(async(() => {
    TestBed.configureTestingModule({
      declarations: [ MapComponent ],
      providers: [ {provide: TopographyService, use: TopographyServiceStub} ]
    })
    .compileComponents();
  }));

  beforeEach(() => {
    fixture = TestBed.createComponent(MapComponent);
    component = fixture.componentInstance;
    fixture.detectChanges();
  });

  it('should create', () => {
    expect(component).toBeTruthy();
  });

  it('changing the basemapHidden attribute should change the basemap in the component\'s MapView', () => {
    component.mapViewReady.then(
      () => {
        component.basemapHidden = true;
        expect(component.basemapHidden).toBe(true);
        expect(component.mapView.map.basemap).toEqual({} as __esri.Basemap);
      }
    );
  });

  it('changing the topographyHidden attribute should change the component\'s topgraphyLayer visibility', () => {
    component.mapViewReady.then(
      () => {
        component.topographyHidden = true;
        expect(component.topographyLayer.visible).toBe(false);
        component.topographyHidden = false;
        expect(component.topographyLayer.visible).toBe(true);
      }
    );
  });

  it('changing the drawingsHidden attribute should change the component\'s drawingsLayer visibility', () => {
    component.mapViewReady.then(
      () => {
        component.drawingsHidden = true;
        expect(component.drawingsLayer.visible).toBe(false);
        component.drawingsHidden = false;
        expect(component.drawingsLayer.visible).toBe(true);
      }
    );
  });

  it('changing the navigationEnabled attribute should activate or deactivate navigation with the map', () => {
    component.mapViewReady.then(
      () => {
        component.navigationEnabled = false;
        expect(component['fixedExtent']).toBeTruthy();
        expect(component['navigationBlockingListeners'].length).toBeGreaterThan(0);

        component.navigationEnabled = true;
        expect(component['fixedExtent']).toBe(undefined);
      }
    );
  });

  it('#setExtent should set the component\'s map extent to a given extent', async () => {
    const [Extent] = await loadModules(['esri/geometry/Extent']);
    const extent: __esri.Extent = new Extent({
      xmin: 780000.0,
      ymin: 5720000.0,
      xmax: 1105000.0,
      ymax: 6100000.0,
      spatialReference: {
          wkid: 102100
      }
    });

    component.mapViewReady.then(
      () => {
        component.setExtent(extent);
        expect(component.mapView.extent.xmin).toBe(780000.0);
        expect(component.mapView.extent.xmax).toBe(1105000.0);
        expect(component.mapView.extent.ymin).toBe(5720000.0);
        expect(component.mapView.extent.ymax).toBe(6100000.0);
      }
    );
  });

  it('#moveMapCenter should move the component\'s map center', () => {
    component.mapViewReady.then(
      () => {
        const mapCenter = component.mapView.center.clone();
        component.moveMapCenter(10.0, 10.0);
        expect(component.mapView.center).not.toEqual(mapCenter);
      }
    );
  });

  it('#modifyMapscale should modify the component\'s map scale', () => {
    component.mapViewReady.then(
      () => {
        const mapScale = component.mapView.scale;
        component.modifyMapScale(10.0);
        expect(component.mapView.scale).not.toEqual(mapScale);
      }
    );
  });

  it('#loadTopography should load the topography graphics for the extent of the map and add them to the component\'s topographyLayer', () => {
    component.mapViewReady.then(
      () => {
        component.loadTopography();
        expect(component.topographyLayer.layers.length).toBe(4);
      }
    );
  });

  it('#setVisibleTopographicCategory should make the topographies belonging to the selected category visible and the other hidden', () => {
    component.mapViewReady.then(
      () => {
        component.loadTopography();

        component.setVisibleTopographicCategory('roads');
        expect(component.topographyLayer.findLayerById(TopographyTypeName.LINEAR).visible).toBe(true);
        expect(component.topographyLayer.findLayerById(TopographyTypeName.WATER).visible).toBe(true);
        expect(component.topographyLayer.findLayerById(TopographyTypeName.BUILDING).visible).toBe(false);
        expect(component.topographyLayer.findLayerById(TopographyTypeName.HARD).visible).toBe(false);

        component.setVisibleTopographicCategory('city');
        expect(component.topographyLayer.findLayerById(TopographyTypeName.LINEAR).visible).toBe(false);
        expect(component.topographyLayer.findLayerById(TopographyTypeName.WATER).visible).toBe(true);
        expect(component.topographyLayer.findLayerById(TopographyTypeName.BUILDING).visible).toBe(true);
        expect(component.topographyLayer.findLayerById(TopographyTypeName.HARD).visible).toBe(true);
      }
    );
  });

  it('#clearLayerGraphics should clear all graphics from the topography and drawings layers and make them hidden', async () => {
    const [
      Graphic,
      GraphicsLayer
    ] = await loadModules([
      'esri/Graphic',
      'esri/layers/GraphicsLayer'
    ]);
    const layer: __esri.GraphicsLayer = new GraphicsLayer();
    component.topographyLayer.add(layer);
    const graphic: __esri.Graphic = new Graphic();
    component.drawingsLayer.add(graphic);

    component.clearLayersGraphics();
    expect(component.topographyLayer.layers.length).toBe(0);
    expect(component.topographyHidden).toBe(true);
    expect(component.drawingsLayer.graphics.length).toBe(0);
    expect(component.drawingsHidden).toBe(true);
  });
});
