import { Component, OnInit, ViewChild, ElementRef } from '@angular/core';
import { loadModules } from 'esri-loader';

import { defaultBasemap, initialExtent, mapsAspectRatio } from './map-defaults';
import { navigationEvents } from './navigation-events';
import { TopographyService } from './topography/topography.service';
import { topographicCategories } from './topography/utils';

@Component({
  selector: 'app-map',
  templateUrl: './map.component.html',
  styleUrls: ['./map.component.css']
})
export class MapComponent implements OnInit {

  @ViewChild('availableSpace') private availableSpace: ElementRef<HTMLDivElement>;
  @ViewChild('mapContainer') private mapContainer: ElementRef<HTMLDivElement>;

  mapView: __esri.MapView;
  mapViewReady: Promise<void>;

  private _basemapHidden = false;
  private basemap: __esri.Basemap;

  private _topographyHidden = true;
  topographyLayer: __esri.GroupLayer;

  private _drawingsHidden = true;
  drawingsLayer: __esri.GraphicsLayer;

  private _resizeOnChange = false;
  private fixedExtent: __esri.Extent | undefined;

  private _navigationEnabled = true;
  private navigationBlockingListeners: IHandle[] = [];

  constructor(
    private topographyService: TopographyService
  ) {}

  ngOnInit() {
    this.mapViewReady = this.init();
  }

  /**
   * Initialise the component's MapView with a default basemap and and initial extent.
   */
  private async init(): Promise<void> {
    const [
      Map,
      MapView,
      GroupLayer,
      GraphicsLayer
    ] = await loadModules([
      'esri/Map',
      'esri/views/MapView',
      'esri/layers/GroupLayer',
      'esri/layers/GraphicsLayer'
    ]);

    const componentMap: __esri.Map = new Map({
      basemap: defaultBasemap
    });

    this.basemap = componentMap.basemap;

    // Initialise the component's MapView.
    this.mapView = new MapView({
      ui: {
        components: ['attribution']
      },
      constraints: {
        minZoom: 8,
        snapToZoom: false
      },
      container: this.mapContainer.nativeElement,
      extent: initialExtent,
      map: componentMap
    });

    this.resizeMapView();

    // Initialise the component's layer to print topographic information on the map.
    this.topographyLayer = new GroupLayer({
      id: 'topography',
      visible: false
    });
    this.mapView.map.add(this.topographyLayer);

    // Initialise the component's layer to print user drawings on the map.
    this.drawingsLayer = new GraphicsLayer({
      id: 'drawings',
      visible: false
    });
    this.mapView.map.add(this.drawingsLayer);

    return this.mapView.when();
  }

  get basemapHidden() {
    return this._basemapHidden;
  }

  set basemapHidden(value: boolean) {
    this._basemapHidden = value;
    this.mapView.map.basemap = !value ? this.basemap : {} as __esri.Basemap;
  }

  get topographyHidden() {
    return this._topographyHidden;
  }

  set topographyHidden(value: boolean) {
    this._topographyHidden = value;

    if (this.topographyLayer) {
      this.topographyLayer.visible = !value;
    }
  }

  get drawingsHidden() {
    return this._drawingsHidden;
  }

  set drawingsHidden(value: boolean) {
    this._drawingsHidden = value;

    if (this.drawingsLayer) {
      this.drawingsLayer.visible = !value;
    }
  }

  get resizeOnChange() {
    return this._resizeOnChange;
  }

  set resizeOnChange(value: boolean) {
    this._resizeOnChange = value;

    if (value) {
      window.addEventListener(
        'resize',
        () => {
          this.resizeMapView();
          if (this.fixedExtent) {
            this.setExtent(this.fixedExtent);
            this.mapView.goTo(this.fixedExtent);
          }
        },
        true
      );
    }
  }

  get navigationEnabled() {
    return this._navigationEnabled;
  }

  set navigationEnabled(value: boolean) {
    if (value) {
      this.fixedExtent = undefined;

      this.navigationBlockingListeners.forEach(
        listener => listener.remove()
      );

    } else if (this._navigationEnabled && !value) {
      this.fixedExtent = this.mapView.extent.clone();

      // Stop any potential current movement on the map.
      this.mapView.goTo(this.fixedExtent);

      this.navigationBlockingListeners = navigationEvents.map(
        (navigationEvent) => this.mapView.on(
          navigationEvent.name,
          navigationEvent.modifiers,
          (event) => event.stopPropagation()
      ));
    }

    this._navigationEnabled = value;
  }

  /**
   * Resize the component's MapView to fit inside its parent view while keeping a given aspect ratio.
   */
  private resizeMapView() {
    // Set the component's container width and height to 0 to prevent
    // it from influencing the availableSpace container's size.
    this.mapContainer.nativeElement.style.width = '0';
    this.mapContainer.nativeElement.style.height = '0';

    const availableWidth = this.availableSpace.nativeElement.clientWidth;
    const availableHeight = this.availableSpace.nativeElement.offsetHeight;

    const availableSpaceRatio = availableWidth / availableHeight;

    let height: number;
    let width: number;
    if (mapsAspectRatio < availableSpaceRatio) {
      height = availableHeight;
      width = height * mapsAspectRatio;
    } else {
      width = availableWidth;
      height = width / mapsAspectRatio;
    }

    width = Math.round(width);
    height = Math.round(height);

    const left = (this.availableSpace.nativeElement.offsetWidth - width) / 2;
    const top = (this.availableSpace.nativeElement.offsetHeight - height) / 2;

    Object.assign(
      this.mapContainer.nativeElement.style,
      {
        height: `${height}px`,
        width: `${width}px`,
        position: 'relative',
        top: `${top}px`,
        left: `${left}px`
      }
    );
  }

  /**
   * Set the extent covered by the component's MapView.
   *
   * @param extent The Extent to be covered by the component's MapView.
   */
  setExtent(extent: __esri.Extent) {
    const extentClone = extent.clone();
    this.mapView.extent = extentClone;
    this.mapView.extent.xmin = extentClone.xmin;
    this.mapView.extent.xmax = extentClone.xmax;
    this.mapView.extent.ymin = extentClone.ymin;
    this.mapView.extent.ymax = extentClone.ymax;
  }

  /**
   * Move the position of the component's MapView center by a given offset.
   *
   * @param x The offset on the x axis to move the MapView's center by.
   * @param y The offset on the y axis to move the MapView's center by.
   */
  moveMapCenter(x: number, y: number): void {
    const mapScreenCenter = this.mapView.toScreen(this.mapView.center);
    mapScreenCenter.x += x;
    mapScreenCenter.y += y;
    this.mapView.goTo(this.mapView.toMap(mapScreenCenter));
  }

  /**
   * Modify the component's MapView scale (zoom level).
   *
   * @param scaleModifier The modification to apply to the MapView's scale.
   */
  modifyMapScale(scaleModifier: number) {
    const percent = this.mapView.scale / 100;
    this.mapView.goTo({
      scale: this.mapView.scale + percent * scaleModifier,
    });
    // this.mapView.scale += percent * scaleModifier;
  }

  /**
   * Load topographic information from the application's ArcGIS server for
   * the current Extent of the component's MapView and add the associated
   * graphics to the component's topographyLayer.
   */
  async loadTopography() {
    this.topographyLayer.removeAll();

    const topographyTypesGraphics = await this.topographyService.loadAllTopographyTypesGraphics(
      this.mapView.extent,
      this.mapView.scale
    );

    this.topographyLayer.addMany(topographyTypesGraphics);
  }

  /**
   * Make the layers for the TopographyTypes belonging to some category visible and
   * hide all the others.
   *
   * @param category The topographic category whose TopographyTypes must be made visible on the map.
   */
  setVisibleTopographicCategory(category: 'roads' | 'city') {
    this.topographyLayer.layers.forEach(
      (layer) => layer.visible = false
    );

    topographicCategories[category].forEach(
      (topographyTypeName) => this.topographyLayer.findLayerById(topographyTypeName).visible = true
    );
  }

  /**
   * Clear all the graphics from the maps topography and drawings layers, and hide them.
   */
  clearLayersGraphics() {
    this.topographyLayer.removeAll();
    this.topographyHidden = true;
    this.drawingsLayer.removeAll();
    this.drawingsHidden = true;
  }
}
