import { loadModules } from 'esri-loader';

/**
 * Interface representing maps in the format saved on the maps server.
 */
export interface ABAMap {
  uid: number;
  title: string;
  city: 0 | 1;
  creationDate: string;
  height?: number;
  width?: number;
  extent?: string;
  graphics?: string;
}

/**
 * Interface representing maps in the format used by the application
 * with the ArcGIS Js API.
 */
export interface EsriMap {
  uid?: number;
  title: string;
  city: 0 | 1;
  height: number;
  width: number;
  extent: __esri.Extent;
  graphics: __esri.Graphic[];
}

/**
 * Transform an ABAMap to an EsriMap.
 *
 * @param abaMap The ABAMap to transform to an EsriMap.
 *
 * @returns An EsriMap containing the transformed data from the input ABAMap.
 */
export async function ABAToEsriMap(abaMap: ABAMap): Promise<EsriMap> {
  const [
    Extent,
    Graphic
  ] = await loadModules([
    'esri/geometry/Extent',
    'esri/Graphic'
  ]);

  return {
    uid: abaMap.uid,
    title: abaMap.title,
    city: abaMap.city,
    height: abaMap.height,
    width: abaMap.width,
    extent: abaMap.extent ? Extent.fromJSON(JSON.parse(abaMap.extent)) : new Extent(),
    graphics: abaMap.graphics ? JSON.parse(abaMap.graphics).map(
      (graphic: any) => Graphic.fromJSON(graphic)
    ) : []
  } as EsriMap;
}
