import { Injectable } from '@angular/core';

import {
  printHeightMm,
  printWidthMm,
  mmToInchMultiplier,
  dpi
} from '../map-defaults';

@Injectable({
  providedIn: 'root'
})
export class PrinterService {

  iFrameElement: HTMLIFrameElement;
  imgElement: HTMLImageElement;
  titleElement: HTMLParagraphElement;
  idElement: HTMLParagraphElement;
  brailleTitleElement: HTMLParagraphElement;
  brailleIdElement: HTMLParagraphElement;

  constructor() {
    if (window.navigator.userAgent.toLowerCase().includes('chrome')) {
      this.initPrintFrame();
    }
  }

  /**
   * Initialize the frame for printing the map.
   */
  private initPrintFrame() {
    this.iFrameElement = document.createElement('iframe');
    document.body.appendChild(this.iFrameElement);

    const style = document.createElement('style');
    style.type = 'text/css';
    style.innerHTML = `
    @page {
      size: landscape A3;
      margin-left: 28mm;
      margin-right: 28mm;
      margin-top: 10mm;
      margin-bottom: 0;
    }

    @font-face {
      font-family: 'braille-custom';
      src: url('/assets/braille.ttf');
    }

    html, body {
      height: 100%;
      width: 100%;
      margin: 0;
      padding: 0;
      text-align: center;
    }

    img {
      border: 2mm solid black;
      margin: 0;
      padding: 0;
    }

    #text-container {
      margin: 2mm 0 0;
      padding: 0;
      width: 100%;
      display: flex;
      justify-content: space-between;
    }

    #text-container p {
      margin: 0;
      padding: 0;
      font-size: 10mm;
      font-family: Arial, Helvetica, sans-serif;
    }

    #braille-container {
      margin: 2mm 0 0;
      padding: 0;
      width: 100%;
      display: flex;
      justify-content: space-between;
    }

    #braille-container p {
      margin: 0;
      padding: 0;
      font-size: 10mm;
      font-family: 'braille-custom';
    }`;

    if (this.iFrameElement.contentDocument) {
      this.iFrameElement.contentDocument.head.appendChild(style);

      this.imgElement = document.createElement('img');

      const textContainer = document.createElement('div');
      textContainer.id = 'text-container';
      this.titleElement = document.createElement('p');
      this.idElement = document.createElement('p');
      textContainer.appendChild(this.titleElement);
      textContainer.appendChild(this.idElement);

      const brailleContainer = document.createElement('div');
      brailleContainer.id = 'braille-container';
      this.brailleTitleElement = document.createElement('p');
      this.brailleIdElement = document.createElement('p');
      brailleContainer.appendChild(this.brailleTitleElement);
      brailleContainer.appendChild(this.brailleIdElement);

      const body = this.iFrameElement.contentDocument.body;
      body.appendChild(this.imgElement);
      body.appendChild(textContainer);
      body.appendChild(brailleContainer);
    }
  }

  /**
   * Print the map inside of a MapView with a title and id.
   *
   * @param id The id of the map to print.
   * @param title The title of the map to print.
   * @param mapView The MapView containing the map to print.
   */
  async printMap(
    id: string,
    title: string,
    mapView: __esri.MapView
  ) {
    // Compute the ideal width and height for the map to print.
    const desiredRatio = (mapView.extent.xmax - mapView.extent.xmin) / (mapView.extent.ymax - mapView.extent.ymin);

    const containerWidth = printWidthMm * mmToInchMultiplier * dpi;
    const containerHeight = printHeightMm * mmToInchMultiplier * dpi;
    const containerAspectRatio = containerWidth / containerHeight;

    let height: number;
    let width: number;

    if (desiredRatio < containerAspectRatio) {
      height = containerHeight;
      width = height * desiredRatio;
    } else {
      width = containerWidth;
      height = width / desiredRatio;
    }

    const screenshot = await mapView.takeScreenshot({
      height: height,
      width: width,
    });
    const imgUrl = screenshot.dataUrl;

    if (window.navigator.userAgent.toLowerCase().includes('chrome')) {
      this.titleElement.innerText = title;
      this.idElement.innerText = id;
      this.brailleTitleElement.innerText = title;
      this.brailleIdElement.innerText = id;

      const imageLoadPromise = new Promise(
        (resolve) => {
          this.imgElement.addEventListener('load', resolve);
          this.imgElement.src = imgUrl;
        }
      );
      await imageLoadPromise;

      setTimeout(() => {
        if (this.iFrameElement.contentWindow) {
          this.iFrameElement.contentWindow.print();
        }
      }, 100);

    } else {
      // In Firefox, the IFrame for printing must be initialised at the moment of
      // printing only (otherwise it'll be empty).
      this.initPrintFrame();
      this.titleElement.innerText = title;
      this.idElement.innerText = id;
      this.brailleTitleElement.innerText = title;
      this.brailleIdElement.innerText = id;

      this.imgElement.src = imgUrl;

      if (this.iFrameElement.contentWindow) {
        this.iFrameElement.contentWindow.print();
      }
    }
  }
}
