import { loadModules } from 'esri-loader';
import { linesImgUrl, dotImgUrl } from 'src/config/urls';

/**
 * The different names of the types of topographic information
 * available on the application's ArcGIS server.
 */
export enum TopographyTypeName {
  BUILDING = 'building',
  HARD = 'hard',
  LINEAR = 'linear',
  WATER = 'water'
}

/**
 * The categories of topographic information types that are
 * grouped together to be drawn on maps.
 */
export const topographicCategories = {
  roads: [
    TopographyTypeName.LINEAR,
    TopographyTypeName.WATER
  ],
  city: [
    TopographyTypeName.BUILDING,
    TopographyTypeName.HARD,
    TopographyTypeName.WATER
  ]
};

/**
 * A type of topographic information which can be drawn on a map.
 * TopographyTypes cover different sub-types of topographic information
 * extracted from the application's ArcGIS server.
 * Each TopographyType has an Esri Symbol associated to it to distringuish
 * it from the others when it is drawn on a map.
 */
export interface TopographyType {
  name: TopographyTypeName;
  topographySubTypes: string[];
  symbol: __esri.Symbol;
}

/**
 * Get the list of all the TopographyTypes for the application.
 */
export async function getAllTopographyTypes(): Promise<TopographyType[]> {
  const [
    SimpleFillSymbol,
    PictureFillSymbol
  ] = await loadModules([
    'esri/symbols/SimpleFillSymbol',
    'esri/symbols/PictureFillSymbol'
  ]);

  return [
    {
      name: TopographyTypeName.BUILDING,
      topographySubTypes: [
        'batiment'
      ],
      symbol: new PictureFillSymbol({
        url: linesImgUrl,
        height: '16px',
        width: '16px'
      })
    },
    {
      name: TopographyTypeName.HARD,
      topographySubTypes: [
        'ilot',
        'trottoir',
        'autre_revetement_dur',
        'rocher',
        'place_aviation',
      ],
      symbol: new SimpleFillSymbol({
        color: '#000000'
      })
    },
    {
      name: TopographyTypeName.LINEAR,
      topographySubTypes: [
        'route_chemin',
        'chemin_de_fer',
      ],
      symbol: new SimpleFillSymbol({
        color: '#000000'
      })
    },
    {
      name: TopographyTypeName.WATER,
      topographySubTypes: [
        'eau_stagnante',
        'bassin',
        'cours_eau',
        'fontaine',
      ],
      symbol: new PictureFillSymbol({
        url: dotImgUrl,
        height: '16px',
        width: '16px'
      })
    }
  ];
}
