import { Injectable } from '@angular/core';
import { saveAs } from 'file-saver';
const geojson = require('geojson');
const tokml = require('tokml');

/**
 * The interface for points in a KML file.
 */
interface KMLPoint {
  name: string;
  lat: number;
  lng: number;
}

@Injectable({
  providedIn: 'root'
})
export class ItineraryService {

  private itineraryPoints: KMLPoint[] = [];

  constructor() { }

  /**
   * Add a point to the service's current itinerary.
   *
   * @param pointName The name of the point to add.
   * @param latitude The latitude of the point to add.
   * @param longitude The longitude of the point to add.
   */
  addPoint(pointName: string, latitude: number, longitude: number) {
    this.itineraryPoints.push({
      name: pointName,
      lat: latitude,
      lng: longitude
    });
  }

  /**
   * Delete the last point in the service's current itinerary.
   *
   * @returns A boolean indicating if the last point could be deleted.
   */
  deleteLastPoint(): boolean {
    if (this.itineraryPoints.length === 0) {
      return false;
    }

    this.itineraryPoints.pop();
    return true;
  }

  /**
   * Clear the service's current itinerary.
   */
  clearItinerary() {
    this.itineraryPoints = [];
  }

  /**
   * Save the service's current itinerary in a KML file.
   *
   * @param filename The name of the file in which the itinerary must be saved.
   *
   * @returns A boolean indicating whether the itinerary could be saved.
   */
  saveItinerary(filename: string): boolean {
    if (this.itineraryPoints.length === 0) {
      return false;
    }

    const geojsonObject = geojson.parse(this.itineraryPoints, {Point: ['lng', 'lat']});
    const kmlNameDescription = tokml(geojsonObject, {
      documentName: 'Itinerary'
    });

    const file = new Blob([kmlNameDescription], {
      type: 'text/kml;charset=utf-9'
    });
    saveAs(file, filename + '.kml');
    return true;
  }
}
