import { Injectable } from '@angular/core';

import { SpeechProvider, WebSpeechProvider } from './speech-providers';
import { TranslateService } from '@ngx-translate/core';
import { translateLanguages, Language, VoiceCommand } from './utils';

@Injectable({
  providedIn: 'root'
})
export class SpeechProcessingService {

  private speechProvider: SpeechProvider;

  constructor(translate: TranslateService) {
    this.speechProvider = new WebSpeechProvider(translateLanguages[translate.currentLang]);
  }

  /**
   * Check whether speech recognition is supported by the browser being used by tha application.
   *
   * @returns A boolean indicating if speech recognition is supported.
   */
  speechRecognitionSupported() {
    return this.speechProvider.speechRecognitionSupported();
  }

  /**
   * Set the speech service's language for speech synthesis and recognition.
   *
   * @param language The language to use for the service.
   */
  setLanguage(language: Language) {
    this.speechProvider.language = language;
  }

  /**
   * Use speech synthesis to say some text.
   *
   * @param text The text to say.
   */
  say(text: string) {
    this.speechProvider.say(text);
  }

  /**
   * Add a voice command to respond to to the service.
   *
   * @param voiceCommand The VoiceCommand to recognise and respond to.
   */
  addCommand(voiceCommand: VoiceCommand) {
    this.speechProvider.addCommand(voiceCommand);
  }

  /**
   * Remove all voice commands from the service.
   */
  removeAllCommands() {
    this.speechProvider.removeAllCommands();
  }
}
