import { Language, VoiceCommand, languageIds } from './utils';

/**
 * The interface of speech providers in the application.
 * SpeechProviders are objects that provide speech synthesis
 * and recognition functionalities.
 */
export interface SpeechProvider {
  language: Language;
  speechRecognitionSupported(): boolean;
  say(text: string): void;
  addCommand(voiceCommand: VoiceCommand): void;
  removeAllCommands(): void;
}

interface AppWindow extends Window {
  [key: string]: any;
}
const appWindow = window as AppWindow;

// For future support in Firefox.
const SpeechRecognition = appWindow['SpeechRecognition'] || appWindow['webkitSpeechRecognition'];

/**
 * A SpeechProvider using the WebSpeech API to provide speech processing functionalities.
 */
export class WebSpeechProvider implements SpeechProvider {

  private _language: Language;

  private speechSynthesis = window.speechSynthesis;
  private pitch = 1;
  private rate = 1;

  private speechRecognition: SpeechRecognition;
  private voiceCommands: VoiceCommand[] = [];
  private currentCommand = '';

  constructor(defaultLanguage: Language) {
    if (SpeechRecognition) {
      this.speechRecognition = new SpeechRecognition();
      this.speechRecognition.continuous = true;

      this.speechRecognition.onresult = (event: SpeechRecognitionEvent) => this.processCommand(event);
      this.speechRecognition.onend = _ => {
        if (this.speechRecognition.continuous) {
          this.speechRecognition.start();
        }
      };
      this.speechRecognition.onspeechend = _ => {
        console.log('Speech stopped being recognised...');
      };

      this.speechRecognition.start();
    } else {
      console.log('SpeechRecognition unavailable in this browser.');
    }

    this.language = defaultLanguage;
  }

  get language() {
    return this._language;
  }

  set language(language: Language) {
    this._language = language;
    if (this.speechRecognition) {
      this.speechRecognition.lang = languageIds[language];
    }
  }

  /**
   * Use speech synthesis to say some input text.
   *
   * @param text The text to say.
   */
  say(text: string) {
    // Interrupt any current speech synthesis.
    this.speechSynthesis.cancel();

    const utterance = new SpeechSynthesisUtterance(text);
    utterance.lang = languageIds[this.language];
    utterance.pitch = this.pitch;
    utterance.rate = this.rate;
    this.speechSynthesis.speak(utterance);
  }

  /**
   * Determine if speech recognition is supported in the browser being used.
   *
   * @returns A boolean value indicating if speech recognition is supported.
   */
  speechRecognitionSupported() {
    return !!this.speechRecognition;
  }

  /**
   * Add a voice command to recognise to the speech recognition.
   *
   * @param voiceCommand The voice command to recognise.
   */
  addCommand(voiceCommand: VoiceCommand) {
    this.voiceCommands.push(voiceCommand);
  }

  /**
   * Remove all voice commands from the speech provider.
   */
  removeAllCommands() {
    this.voiceCommands = [];
  }

  /**
   * Process a SpeechRecognitionEvent triggered by SpeechRecognition:
   * check if the recognised text matches any of the SpeechProvider's commands
   * and execute the associated callback if so.
   *
   * @param event A SpeechRecognitionEvent triggered by SpeechRecognition.
   */
  private processCommand(event: SpeechRecognitionEvent) {
    for (let i = event.resultIndex; i < event.results.length; i++) {
      if (event.results[i].isFinal) {
        this.currentCommand += event.results[i][0].transcript;
      }
    }

    this.currentCommand = this.currentCommand.toLowerCase().trim();
    console.log(this.currentCommand);

    let bestMatch = '';
    let callback: (param: string) => void = (param: string) => {};
    let bestMatchLength = 0;
    this.voiceCommands.forEach(
      (voiceCommand) => {
        voiceCommand.commands.forEach(
          (command) => {
            if (this.currentCommand.indexOf(command) > -1) {
              if (command.length > bestMatchLength) {
                bestMatch = command;
                bestMatchLength = command.length;
                callback = voiceCommand.callback;
              }
            }
          }
        );
      }
    );

    if (bestMatch) {
      // If the current command has no parameters, call it without any.
      if (this.currentCommand[bestMatchLength] === undefined) {
        callback('');

      // If there are parameters, parse them and pass them to the callback.
      // It is up to the callback to decide to use them or not.
      } else if (this.currentCommand[bestMatchLength] === ' ') {
        const param = this.currentCommand.substring(bestMatchLength + 1, this.currentCommand.length);
        callback(param);
      }
    }
    this.currentCommand = '';
  }

}
