import { Component, OnInit, ViewChild, AfterViewInit, OnDestroy } from '@angular/core';
import { Location } from '@angular/common';
import { ActivatedRoute } from '@angular/router';
import { TranslateService } from '@ngx-translate/core';
import { SubscriptionLike } from 'rxjs';
import { loadModules } from 'esri-loader';

import { MapComponent } from '../map/map.component';
import { SpeechProcessingService } from './speech-processing/speech-processing.service';
import { MapsDatabaseService } from '../map/maps-database/maps-database.service';
import { GeographyService } from './geography/geography.service';
import { TransportService } from './transports/transport.service';
import { ItineraryService } from './itinerary/itinerary.service';
import { translateLanguages, recognisedLanguages, Language, languagesTranslate } from './speech-processing/utils';
import { originVector, Plane2d, Vector2d, transformCoordinates } from '../geometry/geometry2d';
import { squaredEuclideanDistance } from './geography/utils';
import { ABAToEsriMap } from '../map/maps-database/utils';

@Component({
  selector: 'app-touchpad',
  templateUrl: './touchpad.component.html',
  styleUrls: ['./touchpad.component.css']
})
export class TouchpadComponent implements OnInit, AfterViewInit, OnDestroy {

  @ViewChild('map') private map: MapComponent;

  private _state: 'reading' | 'searching' | 'itinerary' = 'reading';

  /**
   * The map component's plane.
   */
  private mapPlane: Plane2d = {
    A: originVector,
    B: originVector,
    C: originVector,
    D: originVector
  };
  /**
   * The plane on the touchpad.
   */
  private devicePlane: Plane2d = {
    A: originVector,
    B: originVector,
    C: originVector,
    D: originVector
  };

  private nClicks = 0;
  private _mapCalibrated: boolean;

  private currentMarker: __esri.Graphic;
  private currentPoint: __esri.Point;

  private openedMapGraphics: __esri.Graphic[];

  private searchedPoint: __esri.Point | undefined;

  private locationSubscription: SubscriptionLike;

  constructor(
    private route: ActivatedRoute,
    private location: Location,
    private translate: TranslateService,
    private speechProcessingService: SpeechProcessingService,
    private mapsDBService: MapsDatabaseService,
    private geographyService: GeographyService,
    private transportService: TransportService,
    private itineraryService: ItineraryService
  ) {
    // Set the language of the speech processing service to the lang of the translat service.
    this.speechProcessingService.setLanguage(translateLanguages[this.translate.currentLang]);

    //
    document.onclick = (event: MouseEvent) => {
      // Go to fullscreen whenever the window is clicked.
      const elem = document.getElementsByTagName('body')[0] as any;
      const f = elem.requestFullscreen ||
        elem.msRequestFullscreen ||
        elem.mozRequestFullScreen ||
        elem.webkitRequestFullscreen;
      f.call(elem);

      const click = { x: event.x || event.clientX, y: event.y || event.clientY } as Vector2d;

      if (!this.mapCalibrated) {
        this.calibrateMap(click);
      } else {
        this.handleClick(click);
      }
    };

    // Stop responding to clicks and voice commands when the user navigates
    // back or to another page.
    this.locationSubscription = this.location.subscribe(
     () => {
        document.onclick = () => {};
        // Interrupt the current utterance if there is one being spoken out.
        this.speechProcessingService.say('');
        this.speechProcessingService.removeAllCommands();
      }
    );
  }

  ngOnInit() {
  }

  ngAfterViewInit() {
    // Initialise the touchpad's map with the map saved on the
    // maps database server with the id defined in the URL for the component.
    this.map.mapViewReady.then(
      async () => {
        const id = this.route.snapshot.paramMap.get('id');
        if (!id) {
          return;
        }

        const mapId = +id;
        const abaMap = await this.mapsDBService.getMap(mapId);
        const esriMap = await ABAToEsriMap(abaMap);
        this.map.setExtent(esriMap.extent);
        this.map.navigationEnabled = false;
        this.openedMapGraphics = esriMap.graphics;
      }
    );
  }

  ngOnDestroy() {
    this.locationSubscription.unsubscribe();
  }

  get state() {
    return this._state;
  }

  set state(value: 'reading' | 'searching' | 'itinerary') {
    this._state = value;

    switch (value) {
      case 'reading': {
        this.say('touchpad.voice-instructions.read-mode');
        break;
      }
      case 'searching': {
        this.searchedPoint = undefined;
        break;
      }
      case 'itinerary': {
        this.say('touchpad.voice-instructions.itinerary.start');
        this.itineraryService.clearItinerary();
        break;
      }
    }
  }

  get mapCalibrated() {
    return this.nClicks > 4;
  }

  /**
   * Calibrate the map in the aplpication to match the one on the physical touchpad.
   *
   * @param click A Vector2d representing a click on the map.
   */
  private calibrateMap(click: Vector2d) {
    switch (this.nClicks) {
      case 0:
        this.say('touchpad.voice-instructions.calibration.top-left');
        break;

      case 1:
        const extent = this.map.mapView.extent.clone();
        this.mapPlane.A = { x: extent.xmin, y: extent.ymax } as Vector2d;
        this.mapPlane.B = { x: extent.xmax, y: extent.ymax } as Vector2d;
        this.mapPlane.C = { x: extent.xmin, y: extent.ymin } as Vector2d;
        this.mapPlane.D = { x: extent.xmax, y: extent.ymin } as Vector2d;

        this.devicePlane.A = click;

        this.say('touchpad.voice-instructions.calibration.top-right');
        break;

      case 2:
        this.devicePlane.B = click;
        this.say('touchpad.voice-instructions.calibration.bottom-right');
        break;

      case 3:
        this.devicePlane.D = click;
        this.say('touchpad.voice-instructions.calibration.bottom-left');
        break;

      case 4:
        this.devicePlane.C = click;
        this.say('touchpad.voice-instructions.calibration.complete');
        this.initVoiceCommands();
        break;
    }
    this.nClicks++;
  }

  /**
   * Initialise the voice commands to control the touchpad.
   */
  private initVoiceCommands() {
    const currentLang = this.translate.currentLang;
    const languages = this.translate.getLangs();

    const commandsPromises: Promise<string[]>[] = [];
    languages.forEach(
      (language) => {
        this.translate.use(language);

        // Add the command to change the application's language.
        commandsPromises.push(this.translate.get('touchpad.voice-commands.change-lang').toPromise());
        commandsPromises[commandsPromises.length - 1].then(
          (cmds) => this.speechProcessingService.addCommand({
                      commands: cmds,
                      callback: (param?: string) => param ? this.setLanguage(param) : this.setLanguage('')
                    })
        );

        // Add the command to switch to reading mode.
        commandsPromises.push(this.translate.get('touchpad.voice-commands.read-mode').toPromise());
        commandsPromises[commandsPromises.length - 1].then(
          (cmds) => this.speechProcessingService.addCommand({
                      commands: cmds,
                      callback: () => this.state = 'reading'
                    })
        );

        // Add the command to start searching for an address on the map.
        commandsPromises.push(this.translate.get('touchpad.voice-commands.search').toPromise());
        commandsPromises[commandsPromises.length - 1].then(
          (cmds) => this.speechProcessingService.addCommand({
                      commands: cmds,
                      callback: (param?: string) => param ? this.searchAddress(param) : this.searchAddress('')
                    })
        );

        // Add the command to start searching for the station nearest to the
        // last point pressed on the map.
        commandsPromises.push(this.translate.get('touchpad.voice-commands.transportation-nearest').toPromise());
        commandsPromises[commandsPromises.length - 1].then(
          (cmds) => this.speechProcessingService.addCommand({
                      commands: cmds,
                      callback: () => this.searchNearestStation()
                    })
        );

        // Add the command to start searching for the nearest station that serves a given line.
        commandsPromises.push(this.translate.get('touchpad.voice-commands.transportation-line').toPromise());
        commandsPromises[commandsPromises.length - 1].then(
          (cmds) => this.speechProcessingService.addCommand({
                      commands: cmds,
                      callback: (param?: string) => param ? this.searchNearestStationByLine(param) : this.searchNearestStationByLine('')
                    })
        );

        // Add the command to start searching for the nearest train station.
        commandsPromises.push(this.translate.get('touchpad.voice-commands.train-station').toPromise());
        commandsPromises[commandsPromises.length - 1].then(
          (cmds) => this.speechProcessingService.addCommand({
                      commands: cmds,
                      callback: () => this.searchNearestTrainStation()
                    })
        );

        // Add the command to add a point to an itinerary.
        commandsPromises.push(this.translate.get('touchpad.voice-commands.itinerary-add-point').toPromise());
        commandsPromises[commandsPromises.length - 1].then(
          (cmds) => this.speechProcessingService.addCommand({
                      commands: cmds,
                      callback: (param?: string) => param ? this.addPointToItinerary(param) : this.addPointToItinerary('')
                    })
        );

        // Add the command to delete the latest point in an itinerary.
        commandsPromises.push(this.translate.get('touchpad.voice-commands.itinerary-delete-point').toPromise());
        commandsPromises[commandsPromises.length - 1].then(
          (cmds) => this.speechProcessingService.addCommand({
                      commands: cmds,
                      callback: () => this.deleteLastPointInitinerary()
                    })
        );

        // Add the command to save an itinerary.
        commandsPromises.push(this.translate.get('touchpad.voice-commands.itinerary-save').toPromise());
        commandsPromises[commandsPromises.length - 1].then(
          (cmds) => this.speechProcessingService.addCommand({
                      commands: cmds,
                      callback: (param?: string) => param ? this.saveItinerary(param) : this.saveItinerary('')
                    })
        );
      }
    );

    Promise.all(commandsPromises).then(
      () => {
        this.translate.use(currentLang);
        this.speechProcessingService.setLanguage(translateLanguages[currentLang]);
      }
    );
  }

  /**
   * Respond to a click on the application's map.
   *
   * @param click A Vector2d representing a click on the map.
   */
  private async handleClick(click: Vector2d) {
    const [
      Point,
      webMercatorUtils
    ] = await loadModules([
      'esri/geometry/Point',
      'esri/geometry/support/webMercatorUtils'
    ]);

    // Transform the user's click to a point on the map.
    const clickedPoint = transformCoordinates(click, this.devicePlane, this.mapPlane);
    const mapPoint: __esri.Point = new Point({
      spatialReference: this.map.mapView.spatialReference.clone(),
      x: clickedPoint.x,
      y: clickedPoint.y
    });

    // Inform the user if he clicked outside the map.
    if (!this.isPointOnMap(mapPoint)) {
      this.say('touchpad.voice-instructions.click-outside');
      return;
    }

    // Draw a marker where the user clicked.
    this.drawMarkerOnMap(mapPoint);

    // Transform the map point to WGS84 (geographic) coordinates.
    const geographicPoint: __esri.Point = webMercatorUtils.webMercatorToGeographic(mapPoint);
    this.currentPoint = geographicPoint;

    if (this.state === 'reading') {
      // Check if there are saved points of interest (POIs) near the user click on the map.
      const nearbyPOIs = this.openedMapGraphics.filter(
        (graphic) => graphic.geometry.type === 'point' &&
                     Math.sqrt(squaredEuclideanDistance(graphic.geometry as __esri.Point, mapPoint)) < 10.0
      );

      // If there are any, use speech processing to say the text associated to the first one.
      if (nearbyPOIs.length > 0) {
        this.speechProcessingService.say(nearbyPOIs[0].attributes.text);
        return;
      }

      // If there are no POIs around, get the address of the point.
      this.getAddress(geographicPoint);

    } else if (this.state === 'searching') {
      if (this.searchedPoint) {
        this.getDirections(geographicPoint, this.searchedPoint);

      } else {
        this.say('touchpad.voice-instructions.search.loading');
      }
    }
  }

  /**
   * Set the language for speech processing.
   *
   * @param language The name of the language to use for speech processing.
   */
  private setLanguage(language: string) {
    const lang = recognisedLanguages[language.toLowerCase()];

    if (!lang) {
      this.say('touchpad.voice-instructions.lang-error');
      return;
    }

    this.translate.use(languagesTranslate[lang]);
    this.speechProcessingService.setLanguage(lang);
    this.say('touchpad.voice-instructions.lang-ok');
  }

  /**
   * Check whether a point is on the map or not.
   *
   * @param point An Esri Point with WGS84 coordinates.
   *
   * @returns A boolean indicating if the point is on the map or not.
   */
  private isPointOnMap(point: __esri.Point): boolean {
    // y coordinates on the map are reversed.
    return (point.x > this.mapPlane.A.x && point.x < this.mapPlane.B.x &&
            point.y < this.mapPlane.A.y && point.y > this.mapPlane.C.y);
  }

  /**
   * Draw a marker on the touchpad's map at a given position.
   *
   * @param point An Esri point indicating where the marker must be drawn on the map.
   */
  private async drawMarkerOnMap(point: __esri.Point) {
    const [
      Graphic,
      SimpleMarkerSymbol
    ] = await loadModules([
      'esri/Graphic',
      'esri/symbols/SimpleMarkerSymbol'
    ]);

    // Remove the marker from the previous click on the map.
    if (this.currentMarker) {
      this.map.mapView.graphics.remove(this.currentMarker);
    }

    // Draw a new marker on the map to indicate where the user clicked.
    const markerSymbol: __esri.SimpleMarkerSymbol = new SimpleMarkerSymbol({
      color: [226, 119, 40],
      outline: { color: [255, 255, 255], width: 2 }
    });

    this.currentMarker = new Graphic({
      geometry: point,
      symbol: markerSymbol,
      visible: true
    });

    this.map.mapView.graphics.add(this.currentMarker);
  }

  /**
   * Get the nearest address to a point and use speech synthesis to say it.
   *
   * @param point The point for which the address must be retrieved.
   */
  private async getAddress(point: __esri.Point) {
    const address = await this.geographyService.getAddress(point);
    if (address.length > 0) {
      console.log(address);
      this.speechProcessingService.say(address);
    }
  }

  /**
   * Search for an address, and switch to searching mode if it exists and it is
   * on the map.
   *
   * @param address The address to search.
   */
  private async searchAddress(address: string) {
    this.state = 'searching';
    this.say('touchpad.voice-instructions.search.start', address);

    // Replace 'un' in french by '1' (this is necessary because speech recognition
    // always returns 'un' and never '1').
    const searchAddress = address.replace('un', '1');

    const searchPoint = await this.geographyService.getPoint(searchAddress);

    // If no point was found for the input address, go back to reading mode.
    if (!searchPoint) {
      this.say('touchpad.voice-instructions.search.error');
      // Give time to speech synthesis to say the sentence before switching to
      // reading mode.
      setTimeout(
        () => this.state = 'reading',
        3000
      );
      return;
    }

    // Check that the point for the input address is inside the map. If not, go back
    // to reading mode.
    const [webMercatorUtils] = await loadModules(['esri/geometry/support/webMercatorUtils']);
    const mapPoint: __esri.Point = webMercatorUtils.geographicToWebMercator(searchPoint);
    if (!this.isPointOnMap(mapPoint)) {
      this.say('touchpad.voice-instructions.search.outside');
      setTimeout(
        () => this.state = 'reading',
        3000
      );
      return;
    }

    this.searchedPoint = searchPoint;
  }

  /**
   * Search for the nearest public transportations station. If one is found, start searching
   * for it in searching mode.
   */
  private async searchNearestStation() {
    this.state = 'searching';

    // Check that user pressed a point on the map before searching for the nearest station.
    if (!this.currentPoint) {
      this.say('touchpad.voice-instructions.transportation.no-point-selected');
      setTimeout(
        () => this.state = 'reading',
        4000
      );
      return;
    }

    // Search the station nearest to the last point pressed by the user.
    const nearestStation = await this.transportService.getNearestBusStation(this.currentPoint);

    // If one is found, start searching for it. Otherwise, inform the user that no stations
    // were found nearby.
    if (nearestStation) {
      const [Point] = await loadModules(['esri/geometry/Point']);
      this.searchedPoint = new Point({
        x: nearestStation.coordinate.y,
        y: nearestStation.coordinate.x
      });
      this.say('touchpad.voice-instructions.transportation.start-search-nearest', nearestStation.name);
    } else {
      this.say('touchpad.voice-instructions.transportation.no-station-nearby');
      setTimeout(
        () => this.state = 'reading',
        4000
      );
    }
  }

  /**
   * Search for the nearest station serving a given line. If none is found,
   * go back to reading mode.
   *
   * @param line The line to search for among the nearby stations.
   */
  private async searchNearestStationByLine(line: string) {
    this.state = 'searching';

    if (!this.currentPoint) {
      this.say('touchpad.voice-instructions.transportation.no-point-selected');
      setTimeout(
        () => this.state = 'reading',
        4000
      );
    }

    // Search the station nearest to the last point pressed by the user that serves the input line.
    const nearestStation = await this.transportService.getNearestStationByLine(this.currentPoint, line);

    // If one is found, start searching for it. Otherwise, inform the user that no stations
    // were found nearby.
    if (nearestStation) {
      const [Point] = await loadModules(['esri/geometry/Point']);
      this.searchedPoint = new Point({
        x: nearestStation.coordinate.y,
        y: nearestStation.coordinate.x
      });
      this.say('touchpad.voice-instructions.transportation.start-search-nearest', nearestStation.name);
    } else {
      this.say('touchpad.voice-instructions.transportation.no-stations-nearby');
      setTimeout(
        () => this.state = 'reading',
        4000
      );
    }
  }

  /**
   * Start searching for the nearest train station. If none is found, go back to reading mode.
   */
  private async searchNearestTrainStation() {
    this.state = 'searching';

    if (!this.currentPoint) {
      this.say('touchpad.voice-instructions.transportation.no-point-selected');
      setTimeout(
        () => this.state = 'reading',
        4000
      );
    }

    // Search the train station nearest to the last point pressed by the user.
    const nearestStation = await this.transportService.getNearestTrainStation(this.currentPoint);

    // If one is found, start searching for it. Otherwise, inform the user that no stations
    // were found nearby.
    if (nearestStation) {
      const [Point] = await loadModules(['esri/geometry/Point']);
      this.searchedPoint = new Point({
        x: nearestStation.coordinate.y,
        y: nearestStation.coordinate.x
      });
      this.say('touchpad.voice-instructions.train-station.start-search-nearest', nearestStation.name);
    } else {
      this.say('touchpad.voice-instructions.train-station.no-stations-nearby');
      setTimeout(
        () => this.state = 'reading',
        4000
      );
    }
  }

  /**
   * Get directions from a point A to a point B on the map and use speech synthesis to say
   * the directions.
   *
   * @param pointA The point to compute directions from.
   * @param pointB The point to compute directions to.
   */
  private async getDirections(pointA: __esri.Point, pointB: __esri.Point) {
    const directions = await this.geographyService.getDirections(pointA, pointB);

    // If point A and B are close enough, tell the user he/she has reached his/her destination.
    if (directions.reached) {
      this.say('touchpad.voice-instructions.search.arrived');
      return;
    }

    let message = await this.translate.get('touchpad.voice-instructions.directions.search').toPromise();
    message += ' ' + await this.translate.get('touchpad.voice-instructions.directions.' + directions.direction).toPromise();
    message += ' ' + await this.translate.get('touchpad.voice-instructions.directions.at-distance').toPromise();
    message += ' ' + directions.distance;
    message += ' ' + await this.translate.get('touchpad.voice-instructions.directions.' + directions.unit).toPromise();
    this.speechProcessingService.say(message);
  }

  /**
   * Use speech synthesis to say the text corresponding to some id in the
   * translate service.
   *
   * @param textId The identifier of the text to say in the translate service.
   * @param param Optional additionnal words to append after the text to be said.
   */
  private async say(textId: string, param?: string) {
    let message = await this.translate.get(textId).toPromise();

    if (param) {
      message += ' ' + param;
    }
    this.speechProcessingService.say(message);
  }

  /**
   * Add the last point pressed by the user to the current itinerary in the itinerary
   * service.
   *
   * @param name The name of the point to add.
   */
  private async addPointToItinerary(name: string) {
    // Inform the user when he is trying to add a point to an itinerary while not
    // being in itinerary mode.
    if (this.state !== 'itinerary') {
      this.say('touchpad.voice-instructions.itinerary.wrong-mode');
      return;
    }

    // A point must be selected to be added to the itinerary.
    if (!this.currentPoint) {
      this.say('touchpad.voice-instructions.transportation.no-point-selected');
      return;
    }

    // The user needs to specify a name for the point.
    if (name.length < 1) {
      this.say('touchpad.voice-instructions.itinerary.add-point-error');
      return;
    }

    this.itineraryService.addPoint(name, this.currentPoint.latitude, this.currentPoint.longitude);

    const msg1stHalf = await this.translate.get('touchpad.voice-instructions.itinerary.add-point-1').toPromise();
    const msg2ndHalf = await this.translate.get('touchpad.voice-instructions.itinerary.add-point-2').toPromise();
    this.speechProcessingService.say(msg1stHalf + ' ' + name + ' ' + msg2ndHalf);
  }

  /**
   * Delete the last point in the current itinerary.
   */
  private deleteLastPointInitinerary() {
    if (this.itineraryService.deleteLastPoint()) {
      this.say('touchpad.voice-instructions.itinerary.delete-point');
    } else {
      this.say('touchpad.voice-instructions.itinerary.delete-point-error');
    }
  }

  /**
   * Save the current itinerary in a .kml file.
   *
   * @param name The name of the file in which the itinerary must be saved.
   */
  private saveItinerary(name: string) {
    if (this.state !== 'itinerary') {
      this.say('touchpad.voice-instructions.itinerary.wrong-mode');
      return;
    }

    if (name.length < 1) {
      this.say('touchpad.voice-instructions.itinerary.save-error');
      return;
    }

    if (this.itineraryService.saveItinerary(name)) {
      this.say('touchpad.voice-instructions.itinerary.save', name);
      this.itineraryService.clearItinerary();

    } else {
      this.say('touchpad.voice-instructions.itinerary.save-error');
    }
  }
}
