import { TestBed } from '@angular/core/testing';
import { HttpClientTestingModule, HttpTestingController } from '@angular/common/http/testing';

import { TransportService } from './transport.service';
import { Station } from './utils';
import { loadModules } from 'esri-loader';
import { busStationsSearchUrl } from 'src/config/urls';

describe('TransportService', () => {
  let service: TransportService;
  let httpTestingController: HttpTestingController;
  let point: __esri.Point;

  beforeEach(async () => {
    TestBed.configureTestingModule({
      imports: [
        HttpClientTestingModule
      ]
    });
    service = TestBed.get(TransportService);
    httpTestingController = TestBed.get(HttpTestingController);

    const [Point] = await loadModules(['esri/geometry/Point']);
    point = new Point({
      x: 46.207293,
      y: 6.139643
    });
  });

  it('should be created', () => {
    expect(service).toBeTruthy();
  });

  it('#getNearestBusStation should return undefined when there are no stations around', async () => {
    const resultPromise = service.getNearestBusStation(point);

    const req = httpTestingController.expectOne(busStationsSearchUrl + 'locations?x=' + point.x + '&y=' + point.y + '&type=station');
    expect(req.request.method).toEqual('GET');
    req.flush({stations: []});

    resultPromise.then(
      (result) => expect(result).toBe(undefined)
    );
  });

  it('#getNearestStationByLine should return the nearest station serving a specific line', () => {
    const expectedResult: Station = {
      id: '8592841',
      name: 'Genève, Mercier',
      coordinate: {
        type: 'WGS84',
        x: 46.206884,
        y: 6.139521
      },
      distance: 48
    };

    const resultPromise = service.getNearestStationByLine(point, '15');

    resultPromise.then(
      result => expect(result as Station).toEqual(expectedResult)
    );
  });

  it('#getNearestStationByLine should return undefined when there are no stations with the chosen line around', () => {
    const resultPromise = service.getNearestStationByLine(point, 'A');

    const req = httpTestingController.expectOne(busStationsSearchUrl + 'locations?x=' + point.x + '&y=' + point.y + '&type=station');
    expect(req.request.method).toEqual('GET');
    req.flush({stations: []});

    resultPromise.then(
      (result) => expect(result).toBe(undefined)
    );
  });

  it('#getNearestStationByLine should return undefined when there are no stations around', () => {
    const resultPromise = service.getNearestStationByLine(point, '15');

    const req = httpTestingController.expectOne(busStationsSearchUrl + 'locations?x=' + point.x + '&y=' + point.y + '&type=station');
    expect(req.request.method).toEqual('GET');
    req.flush({stations: []});

    resultPromise.then(
      (result) => expect(result).toBe(undefined)
    );
  });

  it('#getNearestTrainStation should return the nearest train station if there is one', async () => {
    const [Point] = await loadModules(['esri/geometry/Point']);
    point = new Point({
      x: 46.210365,
      y: 6.143007
    });

    const expectedResult: Station = {
      id: '0',
      name: 'Genève',
      coordinate: {
        type: 'WGS84',
        x: 46.21022061021335,
        y: 6.142439281444803
      },
      distance: 0
    };

    const resultPromise = service.getNearestTrainStation(point);

    resultPromise.then(
      result => expect(result as Station).toEqual(expectedResult)
    );
  });

  it('#getNearestTrainStation should return undefined when there are no train stations around', () => {
    const resultPromise = service.getNearestTrainStation(point);

    resultPromise.then(
      result => expect(result).toBe(undefined)
    );
  });
});
